import logging

from melusine.data import load_email_data
from melusine.pipeline import MelusinePipeline


def test_pipeline_debug_logging(caplog):
    with caplog.at_level(logging.DEBUG):
        df = load_email_data()
        pipeline = MelusinePipeline.from_config("demo_pipeline")
        pipeline.transform(df)

        assert "Running transform for Cleaner" in caplog.text
        assert "Running transform for EmergencyDetector" in caplog.text
        assert "Running transform for EmergencyDetector (detect)" in caplog.text
        assert "Running transform for EmergencyDetector (post_detect)" in caplog.text



"""
Contain IO classes implementation.

Contained classes: [IoMixin]
"""
from __future__ import annotations

import logging
from typing import Any, TypeVar

from melusine import config

logger = logging.getLogger(__name__)

T = TypeVar("T", bound="IoMixin")


class InitError(Exception):
    """
    Error raised when object instantiation fails.
    """


class IoMixin:
    """
    Defines generic load methods.
    """

    def __init__(self, **kwargs: Any):
        """Initialize attribute."""
        self.json_exclude_list: list[str] = ["_func", "json_exclude_list"]

    @classmethod
    def from_config(
        cls: type[T],
        config_key: str | None = None,
        config_dict: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> T:
        """
        Instantiate a class from a config key or a config dict.

        Parameters
        ----------
        config_key: str
            Configuration key.
        config_dict: dict[str, Any]
            Dictionary of config.
        kwargs: Any

        Returns
        -------
        _: T
            Instantiated objet.
        """
        # Load from Melusine config
        if config_dict is None:
            if config_key is None:
                raise ValueError("You should specify one and only one of 'config_key' and 'config_value'")
            else:
                config_dict = config[config_key]
        else:
            if config_key is not None:
                raise ValueError("You should specify one and only one of 'config_key' and 'config_value'")

        # Update with keyword arguments
        config_dict.update(**kwargs)

        return cls.from_dict(**config_dict)

    @classmethod
    def from_dict(cls: type[T], **params_dict: dict[str, Any]) -> T:
        """
        Method to instantiate a class based a dict object.

        Parameters
        ----------
        params_dict: dict[str, Any]
            Parameters dict.

        Returns
        -------
        _: T
            Instantiated objet.
        """
        # Exclude parameters starting with an underscore
        init_params = {key: value for key, value in params_dict.items() if not key.startswith("_")}

        try:
            instance = cls(**init_params)
            return instance
        except Exception as error:
            raise InitError(f"Failed to instantiate {cls.__name__} with attributes {init_params}.").with_traceback(
                error.__traceback__
            )



"""
Example script to fit a minimal preprocessing pipeline
"""
import pandas as pd
import pytest

from melusine import config
from melusine.pipeline import MelusinePipeline, PipelineConfigurationError
from melusine.processors import Normalizer, RegexTokenizer


def test_pipeline_basic(dataframe_basic):
    """
    Train a pipeline by explicitly instatiating all the transformers.
    """
    # Input data
    df = dataframe_basic.copy()

    # Instantiate processors
    normalizer = Normalizer(lowercase=True, form="NFKD")
    tokenizer = RegexTokenizer()

    # Create pipeline
    pipe = MelusinePipeline(steps=[("normalizer", normalizer), ("tokenizer", tokenizer)], verbose=True)

    # Fit the pipeline and transform the data
    df_transformed = pipe.transform(df)

    # Most basic test, check that the pipeline returns a pandas DataFrame
    assert isinstance(df_transformed, pd.DataFrame)


@pytest.mark.usefixtures("use_test_config")
def test_pipeline_from_config(dataframe_basic):
    """
    Train a pipeline using transformers defined in a pipeline config file.
    """
    # Input data
    df = dataframe_basic.copy()

    # Set config keys
    normalizer_key = "test_normalizer"
    tokenizer_key = "test_tokenizer"
    pipeline_key = "test_pipeline"

    # Pipeline configuration
    conf_pipeline_basic = {
        "steps": [
            {
                "class_name": "Normalizer",
                "module": "melusine.processors",
                "config_key": normalizer_key,
            },
            {
                "class_name": "RegexTokenizer",
                "module": "melusine.processors",
                "config_key": tokenizer_key,
            },
        ]
    }

    test_conf_dict = config.dict()
    test_conf_dict[pipeline_key] = conf_pipeline_basic
    config.reset(config_dict=test_conf_dict)

    # Create pipeline from a json config file (using config key "my_pipeline")
    pipe = MelusinePipeline.from_config(config_key=pipeline_key, verbose=True)

    # Fit the pipeline and transform the data
    df_transformed = pipe.transform(df)

    # Make basic tests
    assert isinstance(df_transformed, pd.DataFrame)
    assert normalizer_key in pipe.named_steps
    assert tokenizer_key in pipe.named_steps


@pytest.mark.usefixtures("use_test_config")
def test_pipeline_from_dict(dataframe_basic):
    """
    Train a pipeline using transformers defined in a pipeline config file.
    """
    # Input data
    df = dataframe_basic.copy()

    # Set config keys
    normalizer_name = "normalizer"
    tokenizer_key = "test_tokenizer"

    # Pipeline configuration
    conf_pipeline_basic = {
        "steps": [
            {
                "name": normalizer_name,
                "class_name": "Normalizer",
                "module": "melusine.processors",
                "parameters": {
                    "form": "NFKD",
                    "input_columns": ["text"],
                    "lowercase": True,
                    "output_columns": ["text"],
                },
            },
            {
                "class_name": "RegexTokenizer",
                "module": "melusine.processors",
                "config_key": tokenizer_key,
            },
        ]
    }

    # Create pipeline from a json config file (using config key "my_pipeline")
    pipe = MelusinePipeline.from_config(config_dict=conf_pipeline_basic, verbose=True)

    # Fit the pipeline and transform the data
    df_transformed = pipe.transform(df)

    # Make basic tests
    assert isinstance(df_transformed, pd.DataFrame)
    assert normalizer_name in pipe.named_steps
    assert tokenizer_key in pipe.named_steps


@pytest.mark.usefixtures("use_test_config")
def test_missing_config_key():
    """
    Train a pipeline using transformers defined in a pipeline config file.
    """
    # Set config keys
    normalizer_name = "normalizer"

    # Pipeline configuration
    conf_pipeline_basic = {
        "steps": [
            {
                "name": normalizer_name,
                "class_name": "Normalizer",
                "module": "melusine.processors",
                "parameters": {
                    "form": "NFKD",
                    "input_columns": ["text"],
                    "lowercase": True,
                    "output_columns": ["text"],
                },
            },
            {
                "class_name": "RegexTokenizer",
                "module": "melusine.processors",
            },
        ]
    }

    # Create pipeline from a json config file (using config key "my_pipeline")
    with pytest.raises(PipelineConfigurationError):
        _ = MelusinePipeline.from_config(config_dict=conf_pipeline_basic, verbose=True)


@pytest.mark.usefixtures("use_test_config")
def test_invalid_config_key():
    """
    Train a pipeline using transformers defined in a pipeline config file.
    """
    incorrect_config_key = "INCORRECT_CONFIG_KEY"

    # Pipeline configuration
    conf_pipeline_basic = {
        "steps": [
            {
                "class_name": "Normalizer",
                "module": "melusine.processors",
                "config_key": incorrect_config_key,
            },
            {
                "class_name": "RegexTokenizer",
                "module": "melusine.processors",
                "name": "test_name",
                "parameters": {"test_key": "test_value"},
            },
        ]
    }

    # Create pipeline from a json config file (using config key "my_pipeline")
    with pytest.raises(KeyError, match=incorrect_config_key):
        _ = MelusinePipeline.from_config(config_dict=conf_pipeline_basic, verbose=True)


@pytest.mark.usefixtures("use_test_config")
@pytest.mark.parametrize(
    "pipeline_conf",
    [
        pytest.param(
            {
                "NOT_STEPS": [
                    {
                        "class_name": "Normalizer",
                        "module": "test_module",
                        "config_key": "test_key",
                    },
                ]
            },
            id="Missing steps key",
        ),
        pytest.param(
            {
                "steps": [
                    {
                        "class_name": "Normalizer",
                        "config_key": "test_key",
                    },
                ]
            },
            id="Missing module key",
        ),
        pytest.param(
            {
                "steps": [
                    {
                        "class_name": "Normalizer",
                        "config_key": "test_key",
                        "name": "test_name",
                    },
                ]
            },
            id="Missing parameters key",
        ),
        pytest.param(
            {
                "steps": [
                    {
                        "class_name": "Normalizer",
                        "module": "test_module",
                        "name": "test_name",
                        "parameters": "THIS SHOULD BE A DICT",
                    },
                ]
            },
            id="Erroneous parameters type",
        ),
        pytest.param(
            {
                "steps": [
                    {
                        "class_name": "Normalizer",
                        "module": "melusine.processors",
                        "parameters": {
                            "form": "NFKD",
                            "input_columns": ["text"],
                            "lowercase": True,
                            "output_columns": ["text"],
                        },
                    },
                    {
                        "class_name": "RegexTokenizer",
                        "module": "melusine.processors",
                        "config_key": "test_tokenizer",
                    },
                ]
            },
            id="Missing name key",
        ),
    ],
)
def test_pipeline_config_error(pipeline_conf):
    """
    Train a pipeline using transformers defined in a pipeline config file.
    """
    # Create pipeline from a json config file (using config key "my_pipeline")
    with pytest.raises(PipelineConfigurationError):
        _ = MelusinePipeline.from_config(config_dict=pipeline_conf)


def test_missing_input_field(dataframe_basic):
    """
    Try to transform with an ill config pipeline.
    (The tokenizer step expects an input field "my_missing_field" which is not present)
    """
    # Input data
    df = dataframe_basic.copy()

    # Set config keys
    normalizer_name = "normalizer"
    tokenizer_name = "tokenizer"
    missing_field_name = "my_missing_field"

    # Pipeline configuration
    conf_pipeline_basic = {
        "steps": [
            {
                "name": normalizer_name,
                "class_name": "Normalizer",
                "module": "melusine.processors",
                "parameters": {
                    "input_columns": ["text"],
                    "output_columns": ["text"],
                },
            },
            {
                "name": tokenizer_name,
                "class_name": "RegexTokenizer",
                "module": "melusine.processors",
                "parameters": {
                    "input_columns": [missing_field_name],
                    "output_columns": ["tokens"],
                },
            },
        ]
    }

    # Create pipeline from a json config file (using config key "my_pipeline")
    pipe = MelusinePipeline.from_config(config_dict=conf_pipeline_basic, verbose=True)

    # Fit the pipeline and transform the data
    with pytest.raises(ValueError, match=rf"(?s){tokenizer_name}.*{missing_field_name}"):
        pipe.transform(df)



"""
Default : Use default configuration to explore functionalities
Specify a config_dict
Specify a config path
Specify a MELUSINE_CONFIG_DIR environment variable
"""


def from_config():
    # --8<-- [start:from_config]
    from melusine.pipeline import MelusinePipeline

    pipeline = MelusinePipeline.from_config(config_key="demo_pipeline")


# --8<-- [end:from_config]


def from_config_dict():
    # --8<-- [start:from_config_dict]
    from melusine.processors import Normalizer

    normalizer_conf = {
        "input_columns": ["text"],
        "output_columns": ["normalized_text"],
        "form": "NFKD",
        "lowercase": False,
    }

    normalizer = Normalizer.from_config(config_dict=normalizer_conf)


# --8<-- [end:from_config_dict]


def print_config():
    # --8<-- [start:print_config]
    from melusine import config

    print(config["demo_pipeline"])


# --8<-- [end:print_config]


def modify_conf_with_dict():
    # --8<-- [start:modify_conf_with_dict]
    from melusine import config

    # Get a dict of the existing conf
    new_conf = config.dict()

    # Add/Modify a config key
    new_conf["my_conf_key"] = "my_conf_value"

    # Reset Melusine configurations
    config.reset(new_conf)


# --8<-- [end:modify_conf_with_dict]


def modify_conf_with_path():  # pragma: no cover
    """Tested in conf/test_config"""
    # --8<-- [start:modify_conf_with_path]
    from melusine import config

    # Specify the path to a conf folder
    conf_path = "path/to/conf/folder"

    # Reset Melusine configurations
    config.reset(config_path=conf_path)

    # >> Using config_path : path/to/conf/folder


# --8<-- [end:modify_conf_with_path]


def modify_conf_with_env():  # pragma: no cover
    """Tested in conf/test_config"""
    # --8<-- [start:modify_conf_with_env]
    import os

    from melusine import config

    # Specify the MELUSINE_CONFIG_DIR environment variable
    os.environ["MELUSINE_CONFIG_DIR"] = "path/to/conf/folder"

    # Reset Melusine configurations
    config.reset()

    # >> Using config_path from env variable MELUSINE_CONFIG_DIR
    # >> Using config_path : path/to/conf/folder


# --8<-- [end:modify_conf_with_env]


def export_config():  # pragma: no cover
    """Tested in conf/test_config"""
    # --8<-- [start:export_config]
    from melusine import config

    # Specify the path a folder (created if it doesn't exist)
    conf_path = "path/to/conf/folder"

    # Export default configurations to the folder
    files_created = config.export_default_config(path=conf_path)


# --8<-- [end:export_config]



def run():
    # --8<-- [start:simple_pipeline]
    from melusine.data import load_email_data
    from melusine.pipeline import MelusinePipeline

    # Load an email dataset
    df = load_email_data()

    # Load a pipeline
    pipeline = MelusinePipeline.from_config("demo_pipeline")  # (1)!

    # Run the pipeline
    df = pipeline.transform(df)
    # --8<-- [end:simple_pipeline]



def run():
    # --8<-- [start:debug_pipeline]
    from melusine.data import load_email_data
    from melusine.pipeline import MelusinePipeline

    # Load an email dataset
    df = load_email_data()

    # Activate debug mode
    df.debug = True

    # Load the default pipeline
    pipeline = MelusinePipeline.from_config("demo_pipeline")

    # Run the pipeline
    df = pipeline.transform(df)
    # --8<-- [end:debug_pipeline]



"""
Setup tests and import fixtures
"""
import numpy as np
import pytest

from melusine import config

# Declare fixtures
pytest_plugins = [
    "tests.fixtures.backend",
    "tests.fixtures.basic_emails",
    "tests.fixtures.docs",
    "tests.fixtures.pipelines",
    "tests.fixtures.processors",
]


# =============== Generic fixtures ===============
# Print statements inside fixtures are only visible when running
# pytest -s


@pytest.fixture(scope="session")
def df_emails():
    from melusine.data import load_email_data
    from melusine.processors import RegexTokenizer

    # Load data
    df_emails = load_email_data(type="full")

    # Tokenize text
    tokenizer = RegexTokenizer(input_columns="body")
    df_emails = tokenizer.transform(df_emails)

    # Add mock meta features
    df_emails["test_meta__A"] = np.random.randint(0, 2, size=len(df_emails))
    df_emails["test_meta__B"] = np.random.randint(0, 2, size=len(df_emails))

    return df_emails


# =============== Fixtures with "function" scope ===============
@pytest.fixture(scope="function")
def reset_melusine_config():
    """
    When a test modifies the melusine configuration, this fixture can be used to reset the config.
    """
    # Code executed before the test starts
    pass

    # Run the test
    yield

    # Code executed after the test ends
    config.reset()


@pytest.fixture(scope="function")
def use_test_config(conf_normalizer, conf_tokenizer, conf_phraser):
    """
    Add test configurations.
    """
    # Code executed before the test starts
    test_conf_dict = config.dict()

    test_conf_dict["test_tokenizer"] = conf_tokenizer
    test_conf_dict["test_normalizer"] = conf_normalizer
    test_conf_dict["test_phraser"] = conf_phraser

    config.reset(config_dict=test_conf_dict)

    # Run the test
    yield

    # Code executed after the test ends
    config.reset()


@pytest.fixture(scope="function")
def use_dict_backend():
    """
    When a test modifies the melusine backend, this fixture can be used to reset the backend.
    """
    from melusine.backend.active_backend import backend

    # =========== Code executed before the test starts ===========
    # Use a dict backend to test a pipeline message by message
    backend.reset("dict")

    # =========== Run the test ===========
    yield

    # =========== Code executed after the test ends ===========
    backend.reset()



from typing import Dict, List, Optional, Union

import pytest

from melusine.base import MelusineRegex


class VirusRegex(MelusineRegex):
    """
    Detect computer viruses but not software bugs.
    """

    @property
    def positive(self) -> Union[str, Dict[str, str]]:
        return r"virus"

    @property
    def neutral(self) -> Optional[Union[str, Dict[str, str]]]:
        return dict(
            NEUTRAL_MEDICAL_VIRUS="corona virus",
            NEUTRAL_INSECT="ladybug",
        )

    @property
    def negative(self) -> Optional[Union[str, Dict[str, str]]]:
        return dict(
            NEGATIVE_BUG="bug",
        )

    @property
    def match_list(self) -> List[str]:
        return [
            "This email contains a virus",
            "There is a virus in the ladybug software",
            "The corona virus is not a computer virus",
        ]

    @property
    def no_match_list(self) -> List[str]:
        return [
            "This process just had a bug",
            "This is a bug not a virus",
            "There are ladybugs on the windows",
        ]


def test_erroneous_substitution_pattern():
    with pytest.raises(ValueError):
        regex = VirusRegex(substitution_pattern="12345")


def test_method_test():
    regex = VirusRegex()
    regex.test()
    assert True


def test_match_method():
    regex = VirusRegex()
    match_data = regex("The computer virus in the ladybug software caused a bug in the corona virus dashboard")

    assert match_data[MelusineRegex.MATCH_RESULT] is False
    assert match_data[MelusineRegex.POSITIVE_MATCH_FIELD] == {
        "DEFAULT": [{"match_text": "virus", "start": 13, "stop": 18}]
    }
    assert match_data[MelusineRegex.NEUTRAL_MATCH_FIELD] == {
        "NEUTRAL_INSECT": [{"match_text": "ladybug", "start": 26, "stop": 33}],
        "NEUTRAL_MEDICAL_VIRUS": [{"match_text": "corona virus", "start": 63, "stop": 75}],
    }
    assert match_data[MelusineRegex.NEGATIVE_MATCH_FIELD] == {
        "NEGATIVE_BUG": [{"match_text": "bug", "start": 52, "stop": 55}]
    }


def test_direct_match_method():
    regex = VirusRegex()

    bool_match_result = regex.get_match_result("The computer virus")

    assert bool_match_result is True

    bool_match_result = regex.get_match_result(
        "The computer virus in the ladybug software caused a bug in the corona virus dashboard"
    )

    assert bool_match_result is False


def test_describe_method(capfd):
    """
    Test describe method.
    """
    regex = VirusRegex()

    # Negative match on bug (group NEGATIVE_BUG) and ignore ladybug and corona virus
    regex.describe("The computer virus in the ladybug software caused a bug in the corona virus dashboard")
    out, err = capfd.readouterr()
    assert "NEGATIVE_BUG" in out
    assert "start" not in out

    # Same but include match positions
    regex.describe(
        "The computer virus in the ladybug software caused a bug in the corona virus dashboard",
        position=True,
    )
    out, err = capfd.readouterr()
    assert "match result is : NEGATIVE" in out
    assert "NEGATIVE_BUG" in out
    assert "start" in out

    regex.describe("This is a dangerous virus")
    out, err = capfd.readouterr()
    assert "match result is : POSITIVE" in out
    assert "start" not in out

    regex.describe("Nada")
    out, err = capfd.readouterr()
    assert "The input text did not match anything" in out


def test_repr():
    """
    Test __repr__ method
    """
    regex = VirusRegex()
    assert "VirusRegex" in repr(regex)


def test_default_neutral_and_negative():
    """
    Test a regex class using default neutral and negative properties.
    """

    class SomeRegex(MelusineRegex):
        """
        Test class.
        """

        @property
        def positive(self):
            return r"test"

        @property
        def match_list(self):
            return ["test"]

        @property
        def no_match_list(self):
            return ["bip bip"]

    regex = SomeRegex()
    assert regex.neutral is None
    assert regex.negative is None



from typing import Callable, List

import pandas as pd

from melusine.base import BaseMelusineDetector


class MyDetector(BaseMelusineDetector):
    @property
    def transform_methods(self) -> List[Callable]:
        return [self.row_method, self.df_method]

    def row_method(self, row, debug_mode=False):
        input_data = row[self.input_columns[0]]
        row[self.output_columns[0]] = input_data + "_row"
        return row

    def df_method(self, df, debug_mode=False):
        df[self.output_columns[1]] = df[self.input_columns[0]].str.upper() + "_df"
        return df


def test_detector_transform_dataframe_wise():
    df = pd.DataFrame([{"input_col": "test0"}, {"input_col": "test1"}])
    detector = MyDetector(name="test_detector", input_columns=["input_col"], output_columns=["row_output", "df_output"])
    df = detector.transform(df)

    assert df["row_output"].iloc[0] == "test0_row"
    assert df["row_output"].iloc[1] == "test1_row"
    assert df["df_output"].iloc[0] == "TEST0_df"
    assert df["df_output"].iloc[1] == "TEST1_df"



import pytest


@pytest.fixture
def backend_base_data():
    return {
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
    }


def single_input_single_output(value, test_keyword_arg=False):
    return_value = value.upper()
    if test_keyword_arg:
        return_value += "_kwarg"
    return return_value


def single_input_multi_output(value, test_keyword_arg=False):
    return_value = value.upper()
    if test_keyword_arg:
        return_value += "_kwarg"
    return return_value, value.capitalize()


def multi_input_single_output(value1, value2, test_keyword_arg=False):
    return_value = value1
    if test_keyword_arg:
        return_value += "_kwarg"
    return value2 * return_value


def multi_input_multi_output(value1, value2, test_keyword_arg=False):
    return_value = value1.upper()
    if test_keyword_arg:
        return_value += "_kwarg"
    return return_value, value2 * 2


def row_input_single_output(row, test_keyword_arg=False):
    return_value = row["str_col"]
    if test_keyword_arg:
        return_value += "_kwarg"
    return f'{return_value}_{row["int_col"]}'


def row_input_multi_output(row, test_keyword_arg=False):
    return_value = row["str_col"].upper()
    if test_keyword_arg:
        return_value += "_kwarg"
    return return_value, row["int_col"] * 2


def row_input_row_output(row, test_keyword_arg=False):
    return_value = row["str_col"].upper()
    if test_keyword_arg:
        return_value += "_kwarg"
    row["new_str_col"] = return_value
    return row


testcase_single_input_single_output = dict(
    test_name="testcase_single_input_single_output",
    func=single_input_single_output,
    input_columns=["str_col"],
    output_columns=["new_str_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO", "BAR"],
    },
)

testcase_single_input_single_output_kwarg = dict(
    test_name="testcase_single_input_single_output_kwarg",
    func=single_input_single_output,
    input_columns=["str_col"],
    output_columns=["new_str_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO_kwarg", "BAR_kwarg"],
    },
    kwargs=dict(test_keyword_arg=True),
)


testcase_single_input_multi_output = dict(
    test_name="testcase_single_input_multi_output",
    func=single_input_multi_output,
    input_columns=["str_col"],
    output_columns=["new_str_col1", "new_str_col2"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col1": ["FOO", "BAR"],
        "new_str_col2": ["Foo", "Bar"],
    },
)

testcase_single_input_multi_output_kwarg = dict(
    test_name="testcase_single_input_multi_output_kwarg",
    func=single_input_multi_output,
    input_columns=["str_col"],
    output_columns=["new_str_col1", "new_str_col2"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col1": ["FOO_kwarg", "BAR_kwarg"],
        "new_str_col2": ["Foo", "Bar"],
    },
    kwargs=dict(test_keyword_arg=True),
)

testcase_row_input_single_output = dict(
    test_name="testcase_row_input_single_output",
    func=row_input_single_output,
    input_columns=["str_col", "int_col"],
    output_columns=["new_str_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["foo_1", "bar_2"],
    },
)

testcase_row_input_single_output_kwarg = dict(
    test_name="testcase_row_input_single_output_kwarg",
    func=row_input_single_output,
    input_columns=["str_col", "int_col"],
    output_columns=["new_str_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["foo_kwarg_1", "bar_kwarg_2"],
    },
    kwargs=dict(test_keyword_arg=True),
)

testcase_no_input_single_output = dict(
    test_name="testcase_no_input_single_output",
    func=row_input_single_output,
    input_columns=None,
    output_columns=["new_str_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["foo_1", "bar_2"],
    },
)

testcase_no_input_single_output_kwarg = dict(
    test_name="testcase_no_input_single_output_kwarg",
    func=row_input_single_output,
    input_columns=None,
    output_columns=["new_str_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["foo_kwarg_1", "bar_kwarg_2"],
    },
    kwargs=dict(test_keyword_arg=True),
)

testcase_multi_input_no_output = dict(
    test_name="testcase_multi_input_no_output",
    func=row_input_row_output,
    input_columns=["str_col", "int_col"],
    output_columns=None,
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO", "BAR"],
    },
)

testcase_multi_input_no_output_kwarg = dict(
    test_name="testcase_multi_input_no_output_kwarg",
    func=row_input_row_output,
    input_columns=["str_col", "int_col"],
    output_columns=None,
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO_kwarg", "BAR_kwarg"],
    },
    kwargs=dict(test_keyword_arg=True),
)

testcase_row_input_row_output = dict(
    test_name="testcase_row_input_row_output",
    func=row_input_row_output,
    input_columns=None,
    output_columns=None,
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO", "BAR"],
    },
)

testcase_row_input_row_output_kwarg = dict(
    test_name="testcase_row_input_row_output_kwarg",
    func=row_input_row_output,
    input_columns=None,
    output_columns=None,
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO_kwarg", "BAR_kwarg"],
    },
    kwargs=dict(test_keyword_arg=True),
)

testcase_no_input_multi_output = dict(
    test_name="testcase_no_input_multi_output",
    func=row_input_multi_output,
    input_columns=None,
    output_columns=["new_str_col", "new_int_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO", "BAR"],
        "new_int_col": [2, 4],
    },
)

testcase_no_input_multi_output_kwarg = dict(
    test_name="testcase_no_input_multi_output_kwarg",
    func=row_input_multi_output,
    input_columns=None,
    output_columns=["new_str_col", "new_int_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO_kwarg", "BAR_kwarg"],
        "new_int_col": [2, 4],
    },
    kwargs=dict(test_keyword_arg=True),
)

testcase_row_input_multi_output = dict(
    test_name="testcase_row_input_multi_output",
    func=row_input_multi_output,
    input_columns=["str_col", "int_col"],
    output_columns=["new_str_col", "new_int_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO", "BAR"],
        "new_int_col": [2, 4],
    },
)

testcase_row_input_multi_output_kwarg = dict(
    test_name="testcase_row_input_multi_output_kwarg",
    func=row_input_multi_output,
    input_columns=["str_col", "int_col"],
    output_columns=["new_str_col", "new_int_col"],
    expected_data={
        "str_col": ["foo", "bar"],
        "int_col": [1, 2],
        "new_str_col": ["FOO_kwarg", "BAR_kwarg"],
        "new_int_col": [2, 4],
    },
    kwargs=dict(test_keyword_arg=True),
)


testcase_list = [value for key, value in locals().items() if key.startswith("testcase")]


def get_fixture_name(fixture_value):
    return fixture_value.get("test_name", "missing_test_name")


@pytest.fixture(
    params=testcase_list,
    ids=get_fixture_name,
)
def backend_testcase(request):
    """Fixture to test all backend configurations"""
    testcase = request.param
    return testcase



from __future__ import annotations

import sys

from melusine import __version__ as melusine_version


def show_versions() -> None:
    """
    Print out version of melusine and dependencies to stdout.

    Examples
    --------
    >>> melusine.show_versions()  # doctest: +SKIP
    --------Version info---------
    melusine:      3.0.0
    Platform:    Linux-5.15.90.1-microsoft-standard-WSL2-x86_64-with-glibc2.35
    Python:      3.11.3 (main, Apr 15 2023, 14:44:51) [GCC 11.3.0]
    \b
    ----Optional dependencies----
    numpy:       1.24.2
    pandas:      2.0.0
    pytorch:  <not installed>
    """
    # note: we import 'platform' here as a micro-optimisation for initial import
    import platform

    # optional dependencies
    deps = _get_dependency_info()

    # determine key length for alignment
    keylen = max(len(x) for x in [*deps.keys(), "melusine", "Platform", "Python"]) + 1

    print("--------Version info---------")
    print(f"{'melusine:':{keylen}s} {melusine_version}")
    print(f"{'Platform:':{keylen}s} {platform.platform()}")
    print(f"{'Python:':{keylen}s} {sys.version}")

    print("\n----Optional dependencies----")
    for name, v in deps.items():
        print(f"{name:{keylen}s} {v}")


def _get_dependency_info() -> dict[str, str]:
    """
    Collect information about optional dependencies.

    Returns:
        _: Dict of optional dependencies and associated versions.
    """
    # See the list of dependencies in pyproject.toml/setup.cfg
    opt_deps = [
        "tensorflow",
        "torch",
        "torchvision",
        "torchlib",
        "transformers",
    ]
    return {f"{name}:": _get_dependency_version(name) for name in opt_deps}


def _get_dependency_version(dep_name: str) -> str:
    """
    Get the version of a dependency.

    Args:
        dep_name: Name of the dependency.

    Returns:
        _: Dependency version or "<not_installed>"
    """
    # import here to optimize the root melusine import
    import importlib

    try:
        module = importlib.import_module(dep_name)
    except ImportError:
        return "<not installed>"

    if hasattr(module, "__version__"):
        module_version = module.__version__
    else:
        module_version = "<__version__ unavailable>"  # pragma: no cover

    return module_version



def test_tutorial001(add_docs_to_pythonpath):
    from docs_src.Configurations.tutorial001 import (
        from_config,
        from_config_dict,
        modify_conf_with_dict,
        print_config,
    )

    _ = from_config()
    _ = from_config_dict()
    _ = print_config()
    _ = modify_conf_with_dict()



"""
Backend to run transforms on pandas.DataFrame objects.

Implemented classes: [
    PandasBackend,
]
"""
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm import tqdm

from melusine.backend.base_backend import BaseTransformerBackend


class PandasBackend(BaseTransformerBackend):
    """
    Backend class to operate on Pandas DataFrames.
    Inherits from the BaseTransformerBackend abstract class.
    Includes multiprocessing functionalities
    """

    def __init__(self, progress_bar: bool = False, workers: int = 1):
        """
        Parameters
        ----------
        progress_bar: bool
            If True, display progress bar
        workers: int
            Number of workers for multiprocessing
        """
        super().__init__()
        self.progress_bar = progress_bar
        self.workers = workers

    def apply_transform(
        self,
        data: pd.DataFrame,
        func: Callable,
        output_columns: Optional[List[str]] = None,
        input_columns: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Method to apply a transform on a Dataset using the Dict backend.

        Parameters
        ----------
        data: pd.DataFrame
            Data to be transformed
        func: Callable
            Transform function to apply to the input data
        output_columns: List[str]
            List of output columns
        input_columns: List[str]
            List of input columns
        kwargs

        Returns
        -------
        _: pd.DataFrame
            Transformed data
        """
        # Multiprocessing
        if self.workers > 1:
            data = self.apply_transform_multiprocessing(
                data, func, output_columns, input_columns=input_columns, **kwargs
            )
        else:
            data = self.apply_transform_regular(data, func, output_columns, input_columns=input_columns, **kwargs)

        return data

    def apply_transform_regular(
        self,
        data: pd.DataFrame,
        func: Callable,
        output_columns: Optional[List[str]] = None,
        input_columns: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Regular transform (no multiprocessing)

        Parameters
        ----------
        data: pd.DataFrame
            Data to be transformed
        func: Callable
            Transform function to apply to the input data
        output_columns: List[str]
            List of output columns
        input_columns: List[str]
            List of input columns
        kwargs

        Returns
        -------
        _: pd.DataFrame
            Transformed data
        """

        # Setup apply parameters
        expand, new_cols = self.setup_apply_parameters(output_columns)

        # Series apply
        if input_columns and len(input_columns) == 1:
            input_column = input_columns[0]

            result = self.apply_joblib_series(
                s=data[input_column],
                func=func,
                expand=expand,
                progress_bar=self.progress_bar,
                **kwargs,
            )

        # DataFrame apply
        else:
            result = self.apply_joblib_dataframe(
                df=data,
                func=func,
                expand=expand,
                progress_bar=self.progress_bar,
                **kwargs,
            )

        # Collect results
        if not new_cols:
            data = result
        else:
            data[new_cols] = result

        return data

    @staticmethod
    def setup_apply_parameters(
        output_columns: Optional[List[str]] = None,
    ) -> Tuple[Union[None, str], Union[None, str, List[str]]]:
        """
        Parameters
        ----------
        output_columns: List[str]
            List of output columns

        Returns
        -------
        expand: str
        new_cols: Union[None, str, List[str]]
        """
        if not output_columns:
            expand = None
            new_cols: Union[None, str, List[str]] = None
        elif len(output_columns) == 1:
            expand = None
            new_cols = output_columns[0]
        # Multiple output columns
        else:
            expand = "expand"
            new_cols = list(output_columns)

        return expand, new_cols

    def apply_transform_multiprocessing(
        self,
        data: pd.DataFrame,
        func: Callable,
        output_columns: Optional[List[str]] = None,
        input_columns: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Transform with multiprocessing

        Parameters
        ----------
        data: pd.DataFrame
            Data to be transformed
        func: Callable
            Transform function to apply to the input data
        output_columns: List[str]
            List of output columns
        input_columns: List[str]
            List of input columns
        kwargs

        Returns
        -------
        _: pd.DataFrame
            Transformed data
        """
        workers = min(self.workers, int(data.shape[0] // 2))
        workers = max(workers, 1)

        # Dataframe is too small to use multiprocessing
        if workers == 1:
            return self.apply_transform_regular(data, func, output_columns, input_columns=input_columns, **kwargs)

        expand, new_cols = self.setup_apply_parameters(output_columns)

        # Use Series.apply
        if input_columns and len(input_columns) == 1:
            input_column = input_columns[0]
            chunks = Parallel(n_jobs=workers)(
                delayed(self.apply_joblib_series)(
                    s=d[input_column],
                    func=func,
                    expand=expand,
                    progress_bar=self.progress_bar,
                    **kwargs,
                )
                for d in np.array_split(data, workers)
            )

        # Use DataFrame.apply
        else:
            chunks = Parallel(n_jobs=workers)(
                delayed(self.apply_joblib_dataframe)(
                    df=d,
                    func=func,
                    expand=expand,
                    progress_bar=self.progress_bar,
                    **kwargs,
                )
                for d in np.array_split(data, workers)
            )

        if not new_cols:
            data = pd.concat(chunks)
        else:
            data[new_cols] = pd.concat(chunks)

        return data

    @staticmethod
    def apply_joblib_dataframe(
        df: pd.DataFrame,
        func: Callable,
        expand: Optional[str] = None,
        progress_bar: bool = False,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Need to create a function to pass to Joblib Parallel.
        This function can't be a lambda so we need to create a separate function.
        """
        if progress_bar:
            apply_func = "progress_apply"
            tqdm.pandas(desc=func.__name__)
        else:
            apply_func = "apply"

        result = getattr(df, apply_func)(func, axis=1, result_type=expand, **kwargs)

        return result

    @staticmethod
    def apply_joblib_series(
        s: pd.Series,
        func: Callable,
        expand: Optional[str] = None,
        progress_bar: bool = False,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Need to create a function to pass to Joblib Parallel.
        This function can't be a lambda so we need to create a separate function.
        """
        if progress_bar:
            apply_func = "progress_apply"
            tqdm.pandas(desc=func.__name__)
        else:
            apply_func = "apply"

        result = getattr(s, apply_func)(func, **kwargs)
        if expand:
            result = result.apply(pd.Series)

        return result

    def add_fields(self, left: pd.DataFrame, right: pd.DataFrame, fields: Optional[List[str]] = None) -> pd.DataFrame:
        """
        Method to add fields form the right object to the left object.

        Parameters
        ----------
        left: pd.DataFrame
            MelusineDataset object
        right: pd.DataFrame
            Melusine Dataset object
        fields: List[str]
            List of fields to be added

        Returns
        -------
        _: pd.DataFrame
            Left object with added fields
        """
        left[fields] = right[fields]

        return left

    def copy(self, data: pd.DataFrame, fields: Optional[List[str]] = None) -> pd.DataFrame:
        """
        Method to make a copy of the dataset.

        Parameters
        ----------
        data: pd.DataFrame
            MelusineDataset object
        fields: List[str]
            List of fields to include in the copy (by default copy all fields)

        Returns
        -------
        _: pd.DataFrame
            Copy of original object
        """
        if not fields:
            fields = data.columns

        return data[fields].copy()

    def get_fields(self, data: pd.DataFrame) -> List[str]:
        """
        Method to get the list of fields available in the input dataset.

        Parameters
        ----------
        data: pd.DataFrame
            MelusineDataset object

        Returns
        -------
        _: List[str]
            List of dataset fields
        """
        return data.columns.to_list()

    def setup_debug_dict(self, data: pd.DataFrame, dict_name: str) -> pd.DataFrame:
        """
        Method to check if debug_mode is activated.

        Parameters
        ----------
        data: pd.DataFrame
            MelusineDataset object
        dict_name: str
            Name of the debug dict field to be added

        Returns
        -------
        _: pd.DataFrame
            MelusineDataset object
        """
        data[dict_name] = [{} for _ in range(len(data))]

        return data



import pandas as pd
import pytest


@pytest.fixture
def text_basic_hello_world():
    return {"text": "Hello world", "header": "Hello"}


@pytest.fixture
def text_basic_with_accent():
    return {
        "text": "Bonjour\nPouvez-vous résilier mon contrat?\nJean Dupont",
        "header": "Demande de résiliation",
    }


@pytest.fixture
def email_basic_hello_world():
    return {
        "body": "Hello world",
        "header": "Hello",
        "chanel": "mail",
        "from": "jo@gmail.com",
        "to": ["test@maif.fr"],
        "bal": "",
        "date": "",
        "ged_chanel": "",
        "pli_id": "",
        "flux_id": "",
        "entity": "",
        "soc_type": "",
        "soc_num": "",
        "nb_documents": 0,
    }


@pytest.fixture
def email_basic_with_accent():
    return {
        "body": "Bonjour\nPouvez-vous résilier mon contrat?\nJean Dupont",
        "header": "Demande de résiliation",
        "chanel": "mail",
        "from": "jo@gmail.com",
        "to": ["test@maif.fr"],
        "bal": "",
        "date": "",
        "ged_chanel": "",
        "pli_id": "",
        "flux_id": "",
        "entity": "",
        "soc_type": "",
        "soc_num": "",
        "nb_documents": 0,
    }


@pytest.fixture
def dataframe_basic(text_basic_hello_world, text_basic_with_accent):
    return pd.DataFrame(
        [
            text_basic_hello_world,
            text_basic_with_accent,
        ]
    )


@pytest.fixture
def email_dataframe_basic(email_basic_hello_world, email_basic_with_accent):
    return pd.DataFrame(
        [
            email_basic_hello_world,
            email_basic_with_accent,
        ]
    )



"""
Unit tests of the VacationReplyDetector
"""
import pandas as pd
import pytest
from pandas import DataFrame

from melusine.detectors import VacationReplyDetector
from melusine.message import Message
from melusine.pipeline import MelusinePipeline


def test_instanciation():
    """Instanciation base test."""
    detector = VacationReplyDetector(
        name="vacation_reply",
        messages_column="messages",
    )
    assert isinstance(detector, VacationReplyDetector)


@pytest.mark.parametrize(
    "df, good_result",
    [
        (
            DataFrame(
                {
                    "messages": [
                        [
                            Message(
                                text="Bonjour, je vous confirme l'annulation du rdv du 01/01/2022 "
                                + "à 16h. Bien cordialement, John Smith.",
                                tags=[
                                    ("HELLO", "Bonjour,"),
                                    (
                                        "BODY",
                                        "je vous confirme l'annulation du rdv du 01/01/2022 à 16h.",
                                    ),
                                    ("GREETINGS", "Bien cordialement, John Smith."),
                                ],
                            )
                        ]
                    ]
                }
            ),
            False,
        ),
        (
            DataFrame(
                {
                    "messages": [
                        [
                            Message(
                                text="Bonjour, \nActuellement en conge je prendrai connaissance"
                                + " de votre message ulterieurement.\nCordialement,",
                                tags=[
                                    ("HELLO", "Bonjour,"),
                                    (
                                        "BODY",
                                        "Actuellement en conge je prendrai connaissance de votre message ulterieurement.",
                                    ),
                                    ("GREETINGS", "Cordialement, "),
                                ],
                            )
                        ]
                    ]
                }
            ),
            True,
        ),
    ],
)
def test_transform(df, good_result):
    """Unit test of the transform() method."""
    df_copy = df.copy()

    message_column = "messages"

    detector = VacationReplyDetector(
        name="vacation_reply",
        messages_column=message_column,
    )
    output_col = detector.result_column

    df = detector.transform(df)
    result = df[output_col][0]
    assert result == good_result


@pytest.mark.parametrize(
    "df, good_detection_result, good_debug_info",
    [
        (
            DataFrame(
                {
                    "messages": [
                        [
                            Message(
                                text="Bonjour, \nActuellement en conge je prendrai connaissance"
                                + " de votre message ulterieurement.\nCordialement,",
                                tags=[
                                    ("HELLO", "Bonjour,"),
                                    (
                                        "BODY",
                                        "Actuellement en conge je prendrai connaissance de votre message ulterieurement.",
                                    ),
                                    ("GREETINGS", "Cordialement, "),
                                ],
                            )
                        ]
                    ]
                }
            ),
            True,
            {
                "parts": [
                    (
                        "BODY",
                        "Actuellement en conge je prendrai connaissance de votre message ulterieurement.",
                    )
                ],
                "text": "Actuellement en conge je prendrai connaissance de votre message ulterieurement.",
                "VacationReplyRegex": {
                    "match_result": True,
                    "negative_match_data": {},
                    "neutral_match_data": {},
                    "positive_match_data": {
                        "VAC_REP_HOLIDAYS": [{"match_text": "Actuellement " "en " "conge", "start": 0, "stop": 21}],
                        "VAC_REP_OUT_OF_OFFICE": [
                            {"match_text": "je " "prendrai " "connaissance", "start": 22, "stop": 46}
                        ],
                    },
                },
            },
        ),
    ],
)
def test_transform_debug_mode(df, good_detection_result, good_debug_info):
    """Unit test of the debug mode."""
    df_copy = df.copy()

    messages_column = "messages"

    detector = VacationReplyDetector(
        name="vacation_reply",
        messages_column=messages_column,
    )
    output_col = detector.result_column
    debug_dict_col = detector.debug_dict_col

    # Transform data
    df.debug = True
    df = detector.transform(df)

    # Collect results
    result = df[output_col].iloc[0]
    debug_result = df[debug_dict_col].iloc[0]

    # Test result
    assert result == good_detection_result
    assert debug_result == good_debug_info


@pytest.mark.parametrize(
    "df, expected_result",
    [
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": [""],
                    "body": [
                        "Bonjour, \nActuellement en congé je prendrai connaissance"
                        + " de votre message ultérieurement.\nCordialement,"
                    ],
                }
            ),
            True,
        ),
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": [""],
                    "body": [
                        "Bonjour,\nje vous confirme l'annulation du rdv du 01/01/2001 "
                        + "à 16h.\nBien cordialement,\nJohn Smith."
                    ],
                }
            ),
            False,
        ),
    ],
)
def test_pipeline_from_config(df, expected_result):
    """
    Instanciate from a config and test the pipeline.
    """
    # Pipeline config key
    pipeline_key = "vacation_reply_pipeline"

    # Create pipeline from config
    pipeline = MelusinePipeline.from_config(config_key=pipeline_key)

    # Apply pipeline on data
    df_transformed = pipeline.transform(df)
    result = df_transformed["vacation_reply_result"][0]

    # Check
    assert result == expected_result



import pytest

from melusine import config
from melusine.io import IoMixin
from melusine.io._classes import InitError


class FakeClass(IoMixin):
    def __init__(self, test_arg):
        super().__init__()
        self.test_arg = test_arg


def test_from_config_dict():
    config_dict = {"test_arg": "hey"}
    instance = FakeClass.from_config(config_dict=config_dict)

    assert instance.test_arg == "hey"


@pytest.mark.usefixtures("use_dict_backend", "reset_melusine_config")
def test_from_config_key():
    my_dict = {"test_arg": "hey"}
    test_conf_dict = config.dict()
    test_conf_dict["testclass_conf"] = my_dict
    config.reset(config_dict=test_conf_dict)

    instance = FakeClass.from_config(config_key="testclass_conf")

    assert instance.test_arg == "hey"


def test_from_config_dict_error():
    unknown_arg = "unknown_arg"
    config_dict = {"test_arg": "hey", unknown_arg: 42}

    with pytest.raises(InitError, match=f"{FakeClass.__name__}.*{unknown_arg}"):
        _ = FakeClass.from_config(config_dict=config_dict)


def test_from_config_dict_and_config_key_error():
    config_dict = {"test_arg": "hey"}

    with pytest.raises(ValueError):
        _ = FakeClass.from_config(config_dict=config_dict, config_key="blabla")


def test_from_config_dict_and_config_key_none_error():
    with pytest.raises(ValueError):
        _ = FakeClass.from_config()



"""
Classes of detectors.

Implemented classes: [ThanksDetector, VacationReplyDetector, ExpeditorDetector,
ReplyDetector, TransferDetector, RecipientsDetector]

"""
from typing import Any, Dict, List, Tuple

from melusine.base import MelusineDetector, MelusineItem, MelusineRegex
from melusine.message import Message
from melusine.regex import (
    EmergencyRegex,
    ReplyRegex,
    ThanksRegex,
    TransferRegex,
    VacationReplyRegex,
)


class ThanksDetector(MelusineDetector):
    """
    Class to detect emails containing only thanks text.

    Ex:
    Merci à vous,
    Cordialement
    """

    # Class constants
    BODY_PART: str = "BODY"
    THANKS_PART: str = "THANKS"
    GREETINGS_PART: str = "GREETINGS"

    # Intermediate columns
    THANKS_TEXT_COL: str = "thanks_text"
    THANKS_PARTS_COL: str = "thanks_parts"
    HAS_BODY: str = "has_body"
    THANKS_MATCH_COL: str = "thanks_match"

    def __init__(
        self,
        messages_column: str = "messages",
        name: str = "thanks",
    ) -> None:
        """
        Attributes initialization.

        Parameters
        ----------
        messages_column: str
            Name of the column containing the messages.

        name: str
            Name of the detector.
        """

        # Input columns
        self.messages_column = messages_column
        input_columns: List[str] = [self.messages_column]

        # Output columns
        self.result_column = f"{name}_result"
        output_columns: List[str] = [self.result_column]

        # Detection regex
        self.thanks_regex: MelusineRegex = ThanksRegex()

        super().__init__(
            name=name,
            input_columns=input_columns,
            output_columns=output_columns,
        )
        self.complex_regex_key: str

    def pre_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Extract text to analyse.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        # Check if a BODY part is present in the last message
        has_body: bool = row[self.messages_column][0].has_tags(
            target_tags={self.BODY_PART}, stop_at={self.GREETINGS_PART}
        )

        # Extract the THANKS part in the last message
        thanks_parts: List[Tuple[str, str]] = row[self.messages_column][0].extract_parts(target_tags={self.THANKS_PART})

        # Compute THANKS text
        if not thanks_parts:
            thanks_text: str = ""
        else:
            thanks_text = "\n".join(x[1] for x in thanks_parts)

        # Save debug data
        if debug_mode:
            debug_dict = {
                self.THANKS_PARTS_COL: thanks_parts,
                self.THANKS_TEXT_COL: thanks_text,
                self.HAS_BODY: has_body,
            }
            row[self.debug_dict_col].update(debug_dict)

        # Create new columns
        row[self.THANKS_TEXT_COL] = thanks_text
        row[self.HAS_BODY] = has_body

        return row

    def detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Use regex to detect thanks.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        debug_info: Dict[str, Any] = {}

        text: str = row[self.THANKS_TEXT_COL]

        detection_data = self.thanks_regex(text)
        detection_result = detection_data[self.thanks_regex.MATCH_RESULT]

        # Save debug data
        if debug_mode:
            debug_info[self.thanks_regex.regex_name] = detection_data
            row[self.debug_dict_col].update(debug_info)

        # Create new columns
        row[self.THANKS_MATCH_COL] = detection_result

        return row

    def post_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Apply final eligibility rules.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """

        # Match on thanks regex & Does not contain a body
        row[self.result_column] = row[self.THANKS_MATCH_COL] and not row[self.HAS_BODY]

        return row


class VacationReplyDetector(MelusineDetector):
    """
    Implement a detector which detects automatic response message like vacation or out of office replies.
    """

    # Class constants
    CONST_TEXT_COL_NAME: str = "vacation_reply_text"
    CONST_DEBUG_TEXT_KEY: str = "text"
    CONST_DEBUG_PARTS_KEY: str = "parts"

    def __init__(
        self,
        name: str,
        messages_column: str = "messages",
    ) -> None:
        """
        Attributes initialization.

        Parameters
        ----------
        name: str
            Detector's name.
        messages_column: str
            Name of the column containing the messages.
        """
        self.messages_column = messages_column

        # Detection regex
        self.vacation_reply_regex: MelusineRegex = VacationReplyRegex()

        # Input columns
        input_columns: List[str] = [messages_column]

        # Output columns
        self.result_column = f"{name}_result"
        output_columns: List[str] = [self.result_column]

        super().__init__(
            name=name,
            input_columns=input_columns,
            output_columns=output_columns,
        )

    def pre_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Extract/prepare the text to analyse.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        # Last message body
        last_message: Message = row[self.messages_column][0]
        body_parts = last_message.extract_last_body()

        if body_parts:
            row[self.CONST_TEXT_COL_NAME] = "\n".join(text for tag, text in body_parts)
        else:
            row[self.CONST_TEXT_COL_NAME] = ""

        # Prepare and save debug data
        if debug_mode:
            debug_dict: Dict[str, Any] = {
                self.CONST_DEBUG_TEXT_KEY: row[self.CONST_TEXT_COL_NAME],
            }
            if self.messages_column:
                debug_dict[self.CONST_DEBUG_PARTS_KEY] = body_parts
            row[self.debug_dict_col].update(debug_dict)

        return row

    def detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Use regex to detect if an email is an automatic response like an Out of office or Vacation reply.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        debug_info: Dict[str, Any] = {}

        text: str = row[self.CONST_TEXT_COL_NAME]

        detection_data = self.vacation_reply_regex(text)
        detection_result = detection_data[self.vacation_reply_regex.MATCH_RESULT]

        # Save debug data
        if debug_mode:
            debug_info[self.vacation_reply_regex.regex_name] = detection_data
            row[self.debug_dict_col].update(debug_info)

        row[self.result_column] = detection_result

        return row

    def post_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Apply final eligibility rule.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        return row


class ReplyDetector(MelusineDetector):
    """
    The ReplyDetector detects if an email is a reply.

    If the header of the email starts with "re", it returns True.
    If not, it returns False.
    """

    # class constant
    CONST_ANALYSED_TEXT_COL: str = "reply_text"

    def __init__(
        self,
        name: str,
        header_column: str = "clean_header",
    ) -> None:
        """
        Attributes initialization.

        Parameters
        ----------
        name: str
            Name given to the detector.
        header_column: [str]
            Name of the column used for the email header.
        """
        # Set instance attributes
        self.header_column = header_column

        # Detection regex
        self.reply_regex: MelusineRegex = ReplyRegex()

        # Input columns
        input_columns: List[str] = [self.header_column]

        # Output columns
        self.result_column = f"{name}_result"
        output_columns: List[str] = [self.result_column]

        super().__init__(
            name=name,
            input_columns=input_columns,
            output_columns=output_columns,
        )

    def pre_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Retrieve text to analyze.

        Log debug information if debug_mode is True.

        Parameters
        ----------
        row: MelusineItem
            Data of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        # Retrieve text to be analysed
        row[self.CONST_ANALYSED_TEXT_COL] = row[self.header_column].lower()

        # Store debug infos
        if debug_mode:
            debug_dict = {
                self.CONST_ANALYSED_TEXT_COL: row[self.CONST_ANALYSED_TEXT_COL],
            }
            row[self.debug_dict_col].update(debug_dict)

        return row

    def detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Check if a header starts with "RE:".

        Parameters
        ----------
        row: MelusineItem
            Data of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        debug_info: Dict[str, Any] = {}

        text: str = row[self.CONST_ANALYSED_TEXT_COL]

        detection_data = self.reply_regex(text)
        detection_result = detection_data[MelusineRegex.MATCH_RESULT]

        # Save debug data
        if debug_mode:
            debug_info[self.reply_regex.regex_name] = detection_data
            row[self.debug_dict_col].update(debug_info)

        row[self.result_column] = detection_result

        return row

    def post_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Apply final eligibility rules.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        # No implementation needed
        return row


class TransferDetector(MelusineDetector):
    """
    The TransferDetector detects if an email is a transfer.
    It returns True if the header starts with "tr:", "fwd:" of if the meta is not empty.
    """

    # class constant
    CONST_ANALYSED_TEXT_COL: str = "reply_text"

    # Debug columns
    CONST_DEBUG_MESSAGE_META: str = "messages[0].meta"

    def __init__(
        self,
        name: str,
        header_column: str = "clean_header",
        messages_column: str = "messages",
    ) -> None:
        """
        Attributes initialization.

        Parameters
        ----------
        name: str
            Name given to the detector.
        header_column: [str]
            Name of the column used for the email header.
        messages_column: [str]
            Name of the column used for the message.
        """
        # Set instance attributes
        self.header_column = header_column
        self.messages_column = messages_column
        self.transfer_regex: MelusineRegex = TransferRegex()

        # Input columns
        input_columns: List[str] = [self.header_column, self.messages_column]

        # Output columns
        self.result_column = f"{name}_result"
        output_columns: List[str] = [self.result_column]

        super().__init__(
            name=name,
            input_columns=input_columns,
            output_columns=output_columns,
        )

    def pre_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Retrieve text to analyze.

        Log debug information if debug_mode is True.

        Parameters
        ----------
        row: MelusineItem
            Data of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        row[self.CONST_ANALYSED_TEXT_COL] = row[self.header_column].lower()

        # Store debug infos
        if debug_mode:
            debug_dict = {
                self.CONST_ANALYSED_TEXT_COL: row[self.CONST_ANALYSED_TEXT_COL],
            }
            row[self.debug_dict_col].update(debug_dict)

        return row

    def detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Check if a header starts with "tr: , fwd:".
        or if the body begins with metadata (e.g, From: , To:, Subject:, etc.)

        Parameters
        ----------
        row: MelusineItem
            Data of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        debug_info: Dict[str, Any] = {}

        text: str = row[self.CONST_ANALYSED_TEXT_COL]
        meta: str = row[self.messages_column][0].meta

        detection_data = self.transfer_regex(text)
        detection_result = detection_data[MelusineRegex.MATCH_RESULT]

        # Save debug data
        if debug_mode:
            debug_info[self.transfer_regex.regex_name] = detection_data
            debug_info[self.CONST_DEBUG_MESSAGE_META] = meta
            row[self.debug_dict_col].update(debug_info)

        row[self.result_column] = detection_result or meta != ""

        return row

    def post_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Apply final eligibility rules.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        # No implementation needed
        return row


class EmergencyDetector(MelusineDetector):
    """
    Implement a detector which detects automatic response message like vacation or out of office replies.
    """

    # Class constants
    CONST_TEXT_COL_NAME: str = "effective_text"
    CONST_DEBUG_TEXT_KEY: str = "text"

    def __init__(
        self,
        name: str,
        header_column: str = "header",
        text_column: str = "det_normalized_last_body",
    ) -> None:
        """
        Attributes initialization.

        Parameters
        ----------
        name: str
            Detector's name.
        header_column: str
            Name of the column containing the text of the email.
        """
        self.header_column = header_column
        self.text_column = text_column

        # Detection regex
        self.regex: MelusineRegex = EmergencyRegex()

        # Input columns
        input_columns: List[str] = [header_column, text_column]

        # Output columns
        self.result_column = f"{name}_result"
        output_columns: List[str] = [self.result_column]

        super().__init__(
            name=name,
            input_columns=input_columns,
            output_columns=output_columns,
        )

    def pre_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Extract/prepare the text to analyse.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        # Last message body
        message_text: str = row[self.text_column]
        header: str = row[self.header_column]

        row[self.CONST_TEXT_COL_NAME] = "\n".join([header, message_text])

        # Prepare and save debug data
        if debug_mode:
            debug_dict: Dict[str, Any] = {
                self.CONST_DEBUG_TEXT_KEY: row[self.CONST_TEXT_COL_NAME],
            }
            row[self.debug_dict_col].update(debug_dict)

        return row

    def detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Apply regex on the effective text.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        debug_info: Dict[str, Any] = {}

        text: str = row[self.CONST_TEXT_COL_NAME]

        detection_data = self.regex(text)
        detection_result = detection_data[self.regex.MATCH_RESULT]

        # Save debug data
        if debug_mode:
            debug_info[self.regex.regex_name] = detection_data
            row[self.debug_dict_col].update(debug_info)

        row[self.result_column] = detection_result

        return row

    def post_detect(self, row: MelusineItem, debug_mode: bool = False) -> MelusineItem:
        """
        Apply final eligibility rule.

        Parameters
        ----------
        row: MelusineItem
            Content of an email.
        debug_mode: bool
            Debug mode activation flag.

        Returns
        -------
        row: MelusineItem
            Updated row.
        """
        return row



import pytest

from melusine.testing.pipeline_testing import assert_pipeline_results


@pytest.mark.usefixtures("use_dict_backend")
def test_pipeline_testing():
    testcase = {
        "test_name": "Simple test",
        "body": "Hello\r\nGood bye",
        "from": "test@gmail.com",
        "header": "Test header",
        "body_cleaner_expected": {"tmp_clean_body": "Hello\nGood bye"},
    }

    assert_pipeline_results(testcase, "my_pipeline")


@pytest.mark.usefixtures("use_dict_backend")
def test_pipeline_testing_error():
    test_name = "Expected result error"
    pipeline_step_name = "body_cleaner"
    field_name = "tmp_clean_body"
    expected_value = "NotTheRightText"

    testcase = {
        "test_name": test_name,
        "body": "Hello\r\nGood bye",
        "from": "test@gmail.com",
        "header": "Test header",
        f"{pipeline_step_name}_expected": {field_name: expected_value},
    }

    with pytest.raises(AssertionError, match=f"{test_name}.*{pipeline_step_name}.*{field_name}.*{expected_value}"):
        assert_pipeline_results(testcase, "my_pipeline")


@pytest.mark.usefixtures("use_dict_backend")
def test_pipeline_testing_untested_field():
    test_name = "Untested field test"
    pipeline_step_name = "non_existent_step"

    testcase = {
        "test_name": test_name,
        "body": "Hello\r\nGood bye",
        "from": "test@gmail.com",
        "header": "Test header",
        f"{pipeline_step_name}_expected": {"field", "expected_value"},
    }

    with pytest.raises(AssertionError, match=f"{pipeline_step_name}.*{test_name}"):
        assert_pipeline_results(testcase, "my_pipeline")



"""
================================== Test-Case fixtures ==================================
# Overview
Tests-cases are defined in this file and tested using `test_emails_generic`.
The fixture `test_message` will successively take the value of every test-case defined
below.

# Adding a new test case
To add a new test-case, create below a variable named "testcase_${NAME}".
The variable should be a dict with the following fields:
- Input fields (ex: "body", "header", "from", "to", etc):
  Used to create the input email passed to the pipeline
-- Example --
testcase_hello_world = {"body": "Hello World"}

- Expected output fields (ex: normalizer_expected, segmenter_expected, etc):
  Used to verify the data resulting from a transformation
  The `segmenter_expected` field will test the results posterior to the pipeline step
  named `segmenter`.
  The content of an expected output field is a dictionary with a key for each DataFrame
  columns to be tested.
-- Example --
testcase_hello_world = {
  "body": "Hello World",
  "tokenizer_expected: {
    "tokens": ['hello', 'world']"   # <== Test the content of the tokens column
    }
}

# Test Message class attributes:
The `messages` column contains a list of Message class instances.
The value of the class instance attributes can be tested
with the syntax `messages.attribute`.
-- Example --
testcase_hello_world = {
  "body": "Hello World\nMessage transféré\nBonjour Monde",
  "segmenter_expected: {
    "messages.text": ['Hello World', 'Bonjour Monde']"   # <== Test the text attribute
    "messages.meta": [None, 'Message transfere']"   # <== Test the meta attribute
    }
}
========================================================================================
"""
import pytest

testcase_initial_cleaning_1 = dict(
    test_name="Initial leaning line breaks",
    body="BonJour wORLD\r\n L'orem \r\n\t\r\nIp-sum\r  Lo_rem  \nip.sum.",
    body_cleaner_expected={
        "tmp_clean_body": "BonJour wORLD\nL'orem\nIp-sum\nLo_rem\nip.sum.",
    },
    content_tagger_expected={
        "messages.tags": [
            [
                ("HELLO", "BonJour wORLD"),
                ("BODY", "L'orem"),
                ("BODY", "Ip-sum"),
                ("BODY", "Lo_rem"),
                ("BODY", "ip.sum."),
            ],
        ],
    },
    tokenizer_expected={
        "ml_body_tokens": [
            "l",
            "orem",
            "ip-sum",
            "lo_rem",
            "ip",
            "sum",
        ],
    },
)

# Watch-out : Multi-line string are NOT Tuples ("abc" "def") != ("abc", "def")
testcase_initial_cleaning_2 = dict(
    test_name="Initial leaning special characters",
    body="Hello\xa0World\n’œ’    <\nyoo\n>",
    body_cleaner_expected={
        "tmp_clean_body": "Hello World\n'oe' <yoo>",
    },
)


# Watch-out : Multi-line strings are NOT Tuples ("abc" "def") != ("abc", "def")
testcase_segmentation1 = dict(
    test_name="Segmentation Date/Envoyé/À/Objet/Le",
    body=(
        "De : test@free.fr <test@free.fr> \nEnvoyé : lundi 30 août 2021 21:26 \nÀ : _Délégation 00.0 - "
        "Ville <ville@maif.fr> \nObjet : Re: Soc : 0000000P - Votre lettre \nBonjour, \n"
        "Vous trouverez ci-joint l'attestation \nMerci de me confirmer la bonne réception de ce "
        "message. \nVous en remerciant par avance. \nCordialement, \nJean Dupont \nLe 2021-08-18 10:30, "
        "ville@malf.fr a écrit : \nBonjour, \nVeuillez trouver ci-jointe la lettre \nLa visualisation des "
        "fichiers PDF nécessite Adobe Reader. \nSentiments mutualistes. \nLa MAIF \n"
    ),
    segmenter_expected={
        "messages.text": [
            (
                "Bonjour,\nVous trouverez ci-joint l'attestation\nMerci de me confirmer la bonne réception "
                "de ce message.\nVous en remerciant par avance.\nCordialement,\nJean Dupont"
            ),
            (
                "Bonjour,\nVeuillez trouver ci-jointe la lettre\nLa visualisation des fichiers PDF nécessite "
                "Adobe Reader.\nSentiments mutualistes.\nLa MAIF"
            ),
        ],
        "messages.meta": [
            (
                "De : test@free.fr <test@free.fr>\nEnvoyé : lundi 30 août 2021 21:26\nÀ : _Délégation 00.0 - "
                "Ville <ville@maif.fr>\nObjet : Re: Soc : 0000000P - Votre lettre"
            ),
            "Le 2021-08-18 10:30, ville@malf.fr a écrit :",
        ],
    },
    content_tagger_expected={
        "messages.tags": [
            [
                ("HELLO", "Bonjour,"),
                ("BODY", "Vous trouverez ci-joint l'attestation"),
                ("BODY", "Merci de me confirmer la bonne réception de ce message."),
                ("THANKS", "Vous en remerciant par avance."),
                ("GREETINGS", "Cordialement,"),
                ("SIGNATURE_NAME", "Jean Dupont"),
            ],
            [
                ("HELLO", "Bonjour,"),
                ("BODY", "Veuillez trouver ci-jointe la lettre"),
                ("FOOTER", "La visualisation des fichiers PDF nécessite Adobe Reader."),
                ("GREETINGS", "Sentiments mutualistes."),
                ("SIGNATURE_NAME", "La MAIF"),
            ],
        ],
    },
)

# Watch-out : Multi-line string are NOT Tuples ("abc" "def") != ("abc", "def")
testcase_segmentation2 = dict(
    test_name="Segmentation Direct Transfer",
    body=(
        "De : Jean Dupond <jean.dupond.971@gmail.com> \nEnvoyé : mardi 31 août 2021 21:45 \n"
        "À : _Délégation Conseil 000 - La Ville <la.ville000@maif.fr> \n"
        "Objet : Demande d'attestation Identifiant : 0000000N \nBonjour \n"
        "Pouvez-vous me transmettre deux attestations au nom de mes enfants \n- Jane Dupond \n- Joe Dupond \n"
        "Merci par avance \n-- \nCordialement \nMr Jean Dupond"
    ),
    segmenter_expected={
        "messages.text": [
            (
                "Bonjour\nPouvez-vous me transmettre deux attestations au nom de mes enfants\n"
                "- Jane Dupond\n- Joe Dupond\nMerci par avance\n--\nCordialement\nMr Jean Dupond"
            )
        ],
        "messages.meta": [
            (
                "De : Jean Dupond <jean.dupond.971@gmail.com>\nEnvoyé : mardi 31 août 2021 21:45\n"
                "À : _Délégation Conseil 000 - La Ville <la.ville000@maif.fr>\n"
                "Objet : Demande d'attestation Identifiant : 0000000N"
            ),
        ],
    },
    content_tagger_expected={
        "messages.tags": [
            [
                ("HELLO", "Bonjour"),
                (
                    "BODY",
                    "Pouvez-vous me transmettre deux attestations au nom de mes enfants",
                ),
                ("BODY", "- Jane Dupond"),
                ("BODY", "- Joe Dupond"),
                ("THANKS", "Merci par avance"),
                ("GREETINGS", "Cordialement"),
                ("SIGNATURE_NAME", "Mr Jean Dupond"),
            ]
        ],
    },
)

# Watch-out : Multi-line string are NOT Tuples ("abc" "def") != ("abc", "def")
testcase_segmentation3 = dict(
    test_name="Segmentation Début du message transféré",
    body=(
        "Envoyé de mon iPhone \nDébut du message transféré : \nDe: Jane Dupond <jane.dupond@hotmail.fr> \n"
        "Date: 11 août 2021 à 17:04:01 UTC+2 \nÀ: Joe DUPOND <joe@du-pond.fr> \n"
        "Objet: Rép. : X - ETAT DES LIEUX \nBonjour Mme X, \n"
        "Suite à l'état des lieux de ce matin, je suis passé à l'agence transmettre les objets \n"
        "Bien cordialement \nJane Dupond \nEnvoyé de mon iPhone \nLe 11 août 2021à 15:35, "
        "Joe DUPOND <jdupond@gmail.fr> a écrit : \nBonjour, \n"
        "Veuillez trouver ci-joint votre état des lieux sortant. \n"
        "Vous en souhaitant bonne réceptionn, \nBien cordialement, \nJoe DUPOND \n "
    ),
    segmenter_expected={
        "messages.text": [
            "Envoyé de mon iPhone",
            (
                "Bonjour Mme X,\nSuite à l'état des lieux de ce matin, je suis passé à l'agence transmettre "
                "les objets\nBien cordialement\nJane Dupond\nEnvoyé de mon iPhone"
            ),
            (
                "Bonjour,\nVeuillez trouver ci-joint votre état des lieux sortant.\n"
                "Vous en souhaitant bonne réceptionn,\nBien cordialement,\nJoe DUPOND"
            ),
        ],
        "messages.meta": [
            "",
            (
                "Début du message transféré :\nDe: Jane Dupond <jane.dupond@hotmail.fr>\n"
                "Date: 11 août 2021 à 17:04:01 UTC+2\nÀ: Joe DUPOND <joe@du-pond.fr>\n"
                "Objet: Rép. : X - ETAT DES LIEUX"
            ),
            "Le 11 août 2021à 15:35, Joe DUPOND <jdupond@gmail.fr> a écrit :",
        ],
    },
)

# Watch-out : Multi-line string are NOT Tuples ("abc" "def") != ("abc", "def")
testcase_segmentation4 = dict(
    test_name="Segmentation Original Message",
    body=(
        "De : marie@protonmail.com<marie@protonmail.com> \nEnvoyé : mardi 31 août 202114:04 \n"
        "À : _Délégation 00.1- Ville <ville@maif.fr> \n"
        "Objet : Re : Soc : 0000000M - Votre attestation Assurance Habitation Responsabilité civile locative \n"
        "Bonjour, \nJe vous renvoie mon RIB concernant le contrat \nd'assurance Habitation Responsabilité civile "
        "locative. \nBien à vous \nMarie \nN° sociétaire : 0000000M \nSent with Proto_nMail Secure Email. \n"
        "--- Original Message --- \nLe mardi 31 août 2021 à 11:09, <ville@maif.fr> a écrit : \n"
        "Bonjour, \nVeuillez trouver ci-joint l'attestation « Responsabilité civile locative » \n"
        "que vous nous avez demandée. \nLa visualisation des fichiers PDF nécessite Adobe Reader. \n"
        "Sentiments mutualistes. \nLa MAIF \n"
    ),
    segmenter_expected={
        "messages.text": [
            (
                "Bonjour,\nJe vous renvoie mon RIB concernant le contrat\nd'assurance Habitation "
                "Responsabilité civile locative.\nBien à vous\nMarie\nN° sociétaire : 0000000M\n"
                "Sent with Proto_nMail Secure Email."
            ),
            (
                "Bonjour,\nVeuillez trouver ci-joint l'attestation « Responsabilité civile locative »\n"
                "que vous nous avez demandée.\nLa visualisation des fichiers PDF nécessite Adobe Reader.\n"
                "Sentiments mutualistes.\nLa MAIF"
            ),
        ],
        "messages.meta": [
            (
                "De : marie@protonmail.com<marie@protonmail.com>\nEnvoyé : mardi 31 août 202114:04\n"
                "À : _Délégation 00.1- Ville <ville@maif.fr>\nObjet : Re : Soc : 0000000M - Votre attestation "
                "Assurance Habitation Responsabilité civile locative"
            ),
            "Original Message ---\nLe mardi 31 août 2021 à 11:09, <ville@maif.fr> a écrit :",
        ],
    },
)


# Watch-out : Multi-line string are NOT Tuples ("abc" "def") != ("abc", "def")
testcase_segmentation5 = dict(
    test_name="Segmentation Le lun. xxx a écrit",
    body=(
        "Bonjour, \nVeuillez trouver en PJ mon RI \n"
        "Le lun. 30 août 2021à 09:40, DUPOND Marie <marie@maif.fr> a écrit : \nBonjour"
    ),
    segmenter_expected={
        "messages.text": ["Bonjour,\nVeuillez trouver en PJ mon RI", "Bonjour"],
        "messages.meta": [
            "",
            "Le lun. 30 août 2021à 09:40, DUPOND Marie <marie@maif.fr> a écrit :",
        ],
    },
)

# Watch-out : Multi-line string are NOT Tuples ("abc" "def") != ("abc", "def")
testcase_segmentation6 = dict(
    test_name="Segmentation Direct Transfer",
    body="Bonjour et merci\nCordialement",
    content_tagger_expected={
        "messages.tags": [
            [
                ("THANKS", "Bonjour et merci"),
                ("GREETINGS", "Cordialement"),
            ]
        ],
    },
)

# Watch-out : Multi-line string are NOT Tuples ("abc" "def") != ("abc", "def")
testcase_transfer_1 = dict(
    test_name="Direct transfer",
    body="De:\nsender.before.transfer@test.fr\nDate:\n3 mars 2023 à 16:42:50\nBonjour\nbla bla",
    transferred_email_processor_expected={
        "messages.text": ["Bonjour\nbla bla"],
        "messages.meta": ["De:\nsender.before.transfer@test.fr\nDate:\n3 mars 2023 à 16:42:50"],
        "det_original_from": "sender.before.transfer@test.fr",
    },
)

testcase_transfer_2 = dict(
    test_name="FOOTER + Transfer",
    body="Envoyé de mon iphone\nDe:\nsender.before.transfer@test.fr\nDate:\n3 mars 2023 à 16:42:50\nBonjour\nbla bla",
    transferred_email_processor_expected={
        "messages.text": ["Bonjour\nbla bla"],
        "messages.meta": ["De:\nsender.before.transfer@test.fr\nDate:\n3 mars 2023 à 16:42:50"],
        "det_original_from": "sender.before.transfer@test.fr",
    },
)

testcase_transfer_3 = {
    "test_name": "FOOTER + Transfer (no email address in meta)",
    "from": "email_sender@test.fr",
    "body": "Envoyé de mon iphone\nDe:\nJohn Doe\nDate:\n3 mars 2023 à 16:42:50\nBonjour\nbla bla",
    "transferred_email_processor_expected": {
        "messages.text": ["Bonjour\nbla bla"],
        "messages.meta": ["De:\nJohn Doe\nDate:\n3 mars 2023 à 16:42:50"],
        "det_original_from": None,
    },
}

testcase_transfer_4 = {
    "test_name": "BODY + Transfer",
    "from": "email_sender@test.fr",
    "body": "Ceci est un BODY\nDe:\nsender.before.transfer@test.fr\nDate:\n3 mars 2023 à 16:42:50\nBonjour\nbla bla",
    "transferred_email_processor_expected": {
        "messages.text": ["Ceci est un BODY", "Bonjour\nbla bla"],
        "messages.meta": ["", "De:\nsender.before.transfer@test.fr\nDate:\n3 mars 2023 à 16:42:50"],
        "det_original_from": None,
    },
}

testcase_false_thanks = dict(
    test_name="Thanks with question mark",
    body="Bonjour\nQu'en est-il svp ? Merci\nJoe Dupont",
    thanks_detector_expected={
        "thanks_result": False,
    },
)

testcase_basic_thanks = dict(
    test_name="Basic Thanks",
    body="Bonjour\nMerci pour cette réponse\nSincèrement\nBLA Bla BLA",
    thanks_detector_expected={
        "thanks_result": True,
    },
)


testcase_false_vacation_reply = dict(
    test_name="Simple vacation reply (False)",
    body="Bonjour\nQu'en est-il svp ? Merci\nJoe Dupont",
    vacation_reply_detector_expected={
        "vacation_reply_result": False,
    },
)

testcase_true_vacation_reply = dict(
    test_name="Simple vacation reply (True)",
    body="Bonjour, \nActuellement en congé je prendrai connaissance"
    + " de votre message ultérieurement.\nCordialement,",
    vacation_reply_detector_expected={
        "vacation_reply_result": True,
    },
)


testcase_real_message = dict(
    test_name="real_email_1",
    body=(
        "De :\n_Délégation - Ville <ma-ville@maif.fr>\n\n\nEnvoyé :\nlundi 27 septembre 2021 22:19\n\n\n"
        "À :\nTEST <test@maif.fr>\n\n\nObjet :\nTR : Soc : 1111111A - Votre"
        " attestation assurance\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nDe :\nJean Dupont <jdupont@gmail.com>"
        "\n\n\nEnvoyé :\nlundi 27 septembre 2021 22:23:43 (UTC+01:00) Brussels, Copenhagen, Madrid, Paris"
        "\n\n\nÀ :\n_Délégation - Ville <ma-ville06@maif.fr>\n\n\nSujet :\nRE: Soc : 1111111A - "
        "Votre attestation assurance\n\n\n\n\n\n\n\n\n\n\n\n\n\nBonjour,\n\n\n\n\nJe vous remercie"
        "pour votre attestation.\n\n\n\n\nLe notaire chante\n\n\n\n\n\n\n\n\n\n\nCordialement,\n\n\n\n\n"
        "Jean Dupont\n\n\n\n\n\n\n\n\n\n\n\nDe :\nma-ville@maif.fr <ma-ville@maif.fr>\n\n\n"
        "Envoyé :\nvendredi 24 septembre 2021 17:42\n\n\nÀ :\njdupont@gmail.com\n\n\nObjet :\n"
        "Soc : 1111111A - Votre attestation assurance\n\n\n\n\n\n\n\n\n\nBonjour,\n\n"
        "Veuillez trouver ci-joint l'attestation.\n\nLa visualisation des fichiers PDF nécessite "
        "Adobe Reader.\n\nSentiments mutualistes.\n\nLa MAIF"
    ),
    messages=[
        "",
        (
            "Bonjour,\n\n\n\n\nJe vous remercie pour votre attestation.\n\n\n\n\n"
            "Le notaire chante\n\n\n\n\n\n\n\n\n\n\nCordialement,\n\n\n\n\nJean Dupont"
        ),
        (
            "Bonjour,\n\nVeuillez trouver ci-joint l'attestation.\n\nLa visualisation des fichiers PDF "
            "nécessite Adobe Reader.\n\nSentiments mutualistes.\n\nLa MAIF"
        ),
    ],
)


testcase_true_reply = dict(
    test_name="Replydetecteur (True)",
    header="Re: Suivi de dossier",
    reply_detector_expected={
        "reply_result": True,
    },
)

testcase_true_reply1 = dict(
    test_name="Replydetecteur (True)",
    header="re: Suivi de dossier",
    reply_detector_expected={
        "reply_result": True,
    },
)
testcase_false_reply = dict(
    test_name="Replydetecteur (false)",
    header="tr: Suivi de dossier",
    reply_detector_expected={
        "reply_result": False,
    },
)
testcase_false_reply1 = dict(
    test_name="Replydetecteur (false)",
    header="",
    reply_detector_expected={
        "reply_result": False,
    },
)
testcase_true_transfer = dict(
    test_name="Transferdetecteur (True)",
    header="Tr: Suivi de dossier",
    body="Bonjour,\n\n\n\n\n\nUn taux d’humidité de 30% a été relevé le 01/01/2024.\n\n\n\nNous reprendrons contact avec l’assurée"
    + " en Janvier 2024.\n\n\n\n\n\n\nBien cordialement,\n\n\n\n\n\nNuméro Auxiliaire : 000000A / 000000B\n\n\n\n\n\n\n\n\n"
    + "Smith KIM\n\n\nTEST\n-\n\nL\na\nV\nalorisation du\nP\natrimoine\n\n\n2, rue du Test\n\n\n00000 Niort"
    + "\n\n\n\n\n\n\n\nTél :    0101010101\n\n\nPort :  0101010101\n\n\nhttp://test.fr\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
    + "\n\n\nDe :\nAccueil - Alex Dupond <alex@test.fr>\n\n\n\nEnvoyé :\njeudi 01 janvier 2024 01:01\n\n\nÀ "
    + ":\nCommercial <etudes@test.fr>\n\n\nObjet :\nTR: Evt : A000000000B survenu le 15/10/2021 - Intervention entreprise"
    + " partenaire\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nDe :\n\n\ntest@maif.fr\n[\nmailto:test@maif.fr\n]\n\n\n\n"
    + "Envoyé :\njeudi 01 janvier 2024 01:01\n\n\nÀ :\nAccueil - Alex Dupond\n\n\nObjet :\nEvt : A000000000B survenu le 01/01/2024"
    + " - Intervention entreprise partenaire\n\n\n\n\n\nMerci de bien vouloir prendre connaissance du document ci-joint.",
    transfer_detector_expected={
        "transfer_result": True,
    },
)
testcase_true_transfer1 = dict(
    test_name="Transferdetecteur (True)",
    header="Suivi de dossier",
    body="De :\nAccueil - Alex Dupond <accueil@test.fr>\n\n\n\nEnvoyé :\njeudi 01 janvier 2024 01:01\n\n\nÀ "
    + ":\nCommercial <etudes@test.fr>\n\n\nObjet :\nTR: Evt : A000000000B survenu le 15/10/2021 - Intervention entreprise"
    + " partenaire\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nDe :\n\n\ntest@maif.fr\n[\nmailto:test@maif.fr\n]\n\n\n\n"
    + "Envoyé :\njeudi 01 janvier 2024 01:01\n\n\nÀ :\nAccueil - Alex Dupond\n\n\nObjet :\nEvt : A000000000B survenu le 01/01/2024"
    + " - Intervention entreprise partenaire\n\n\n\n\n\nMerci de bien vouloir prendre connaissance du document ci-joint.",
    transfer_detector_expected={
        "transfer_result": True,
    },
)
testcase_false_transfer = dict(
    test_name="Transferdetecteur (False)",
    header="test",
    body="Bonjour, ceci est un message de test",
    transfer_detector_expected={
        "transfer_result": False,
    },
)
testcase_false_transfer1 = dict(
    test_name="Transferdetecteur (False)",
    header="",
    body="",
    transfer_detector_expected={
        "transfer_result": False,
    },
)

testcase_list = [value for key, value in locals().items() if key.startswith("testcase")]


def get_fixture_name(fixture_value):
    return fixture_value.get("test_name", "missing_test_name")


@pytest.fixture(
    params=testcase_list,
    ids=get_fixture_name,
)
def testcase(request, default_pipeline="my_pipeline"):
    testcase = request.param

    # Set default testcase parameters
    testcase["pipeline"] = testcase.get("pipeline", default_pipeline)

    # Set default email fields
    testcase["body"] = testcase.get("body", "")
    testcase["header"] = testcase.get("header", "")
    testcase["from"] = testcase.get("from", "")
    testcase["to"] = testcase.get("to", "")
    testcase["attachments"] = testcase.get("attachments", list())

    return testcase



from pathlib import Path

import pandas as pd


def load_email_data() -> pd.DataFrame:
    """
    Function to load a file containing toy email data.

    Return
    ------
    pandas.DataFrame
        DataFrame with toy email data
    """

    # Path to data directory
    data_directory = Path(__file__).parent.resolve()

    # Load data
    email_data_path = data_directory / "emails.json"
    df = pd.read_json(email_data_path, orient="records").fillna("")

    return df



from typing import List

# --8<-- [start:create_dataset]
import pandas as pd
from transformers import pipeline

from melusine.base import MelusineDetector


def create_dataset():
    df = pd.DataFrame(
        [
            {
                "header": "Dossier 123456",
                "body": "Merci beaucoup pour votre gentillesse et votre écoute !",
            },
            {
                "header": "Réclamation (Dossier 987654)",
                "body": ("Bonjour, je ne suis pas satisfait de cette situation, " "répondez-moi rapidement svp!"),
            },
        ]
    )

    return df


# --8<-- [end:create_dataset]


def transformers_standalone():
    # --8<-- [start:transformers]
    model_name_or_path = "cmarkea/distilcamembert-base-nli"

    sentences = [
        "Quelle belle journée aujourd'hui",
        "La marée est haute",
        "Ce film est une catastrophe, je suis en colère",
    ]

    classifier = pipeline(task="zero-shot-classification", model=model_name_or_path, tokenizer=model_name_or_path)

    result = classifier(
        sequences=sentences, candidate_labels=", ".join(["positif", "négatif"]), hypothesis_template="Ce texte est {}."
    )
    # --8<-- [end:transformers]

    return result


# --8<-- [start:detector_init]
class DissatisfactionDetector(MelusineDetector):
    """
    Detect if the text expresses dissatisfaction.
    """

    # Dataframe column names
    OUTPUT_RESULT_COLUMN = "dissatisfaction_result"
    TMP_DETECTION_INPUT_COLUMN = "detection_input"
    TMP_DETECTION_OUTPUT_COLUMN = "detection_output"

    # Model inference parameters
    POSITIVE_LABEL = "positif"
    NEGATIVE_LABEL = "négatif"
    HYPOTHESIS_TEMPLATE = "Ce texte est {}."

    def __init__(self, model_name_or_path: str, text_columns: List[str], threshold: float):
        self.text_columns = text_columns
        self.threshold = threshold
        self.classifier = pipeline(
            task="zero-shot-classification", model=model_name_or_path, tokenizer=model_name_or_path
        )

        super().__init__(input_columns=text_columns, output_columns=[self.OUTPUT_RESULT_COLUMN], name="dissatisfaction")
        # --8<-- [end:detector_init]

    # --8<-- [start:pre_detect]
    def pre_detect(self, row, debug_mode=False):
        # Assemble the text columns into a single text
        effective_text = ""
        for col in self.text_columns:
            effective_text += "\n" + row[col]
        row[self.TMP_DETECTION_INPUT_COLUMN] = effective_text

        # Store the effective detection text in the debug data
        if debug_mode:
            row[self.debug_dict_col] = {"detection_input": row[self.TMP_DETECTION_INPUT_COLUMN]}

        return row
        # --8<-- [end:pre_detect]

    # --8<-- [start:detect]
    def detect(self, row, debug_mode=False):
        # Run the classifier on the text
        pipeline_result = self.classifier(
            sequences=row[self.TMP_DETECTION_INPUT_COLUMN],
            candidate_labels=", ".join([self.POSITIVE_LABEL, self.NEGATIVE_LABEL]),
            hypothesis_template=self.HYPOTHESIS_TEMPLATE,
        )
        # Format classification result
        result_dict = dict(zip(pipeline_result["labels"], pipeline_result["scores"]))
        row[self.TMP_DETECTION_OUTPUT_COLUMN] = result_dict

        # Store ML results in the debug data
        if debug_mode:
            row[self.debug_dict_col].update(result_dict)

        return row
        # --8<-- [end:detect]

    # --8<-- [start:post_detect]
    def post_detect(self, row, debug_mode=False):
        # Compare classification score to the detection threshold
        if row[self.TMP_DETECTION_OUTPUT_COLUMN][self.NEGATIVE_LABEL] > self.threshold:
            row[self.OUTPUT_RESULT_COLUMN] = True
        else:
            row[self.OUTPUT_RESULT_COLUMN] = False

        return row

    # --8<-- [end:post_detect]


def run():
    # --8<-- [start:run]
    df = create_dataset()

    detector = DissatisfactionDetector(
        model_name_or_path="cmarkea/distilcamembert-base-nli",
        text_columns=["header", "body"],
        threshold=0.7,
    )

    df = detector.transform(df)
    # --8<-- [end:run]

    # Debug mode
    df = create_dataset()
    df.debug = True
    _ = detector.transform(df)

    return df


if __name__ == "__main__":  # pragma no cover
    run()



"""
Unit tests of the ReplyDetector.
"""
import pandas as pd
import pytest
from pandas import DataFrame

from melusine.detectors import ReplyDetector
from melusine.pipeline import MelusinePipeline


def test_instantiation():
    """Instanciation base test."""

    # Instantiate manually a detector
    detector = ReplyDetector(
        name="reply",
        header_column="clean_header",
    )
    assert isinstance(detector, ReplyDetector)


@pytest.mark.parametrize(
    "row, good_result",
    [
        (
            {"reply_text": "Devis habitation"},
            False,
        ),
        (
            {"reply_text": "tr: Devis habitation"},
            False,
        ),
        (
            {"reply_text": "re: Envoi d'un document de la Société Imaginaire"},
            True,
        ),
        (
            {"reply_text": "re : Virement"},
            True,
        ),
        (
            {"reply_text": ""},
            False,
        ),
    ],
)
def test_deterministic_detect(row, good_result):
    """Method base test."""

    # Instanciate manually a detector
    detector = ReplyDetector(
        name="reply",
        header_column="clean_header",
    )
    # Test method
    row = detector.detect(row)
    res = row[detector.result_column]
    assert res == good_result


@pytest.mark.parametrize(
    "df_emails, expected_result",
    [
        (
            DataFrame(
                {
                    "clean_header": ["Re: Suivi de dossier"],
                }
            ),
            True,
        ),
        (
            DataFrame(
                {
                    "clean_header": ["Suivi de dossier"],
                }
            ),
            False,
        ),
        (
            DataFrame(
                {
                    "clean_header": ["Tr: Suivi de dossier"],
                }
            ),
            False,
        ),
        (
            DataFrame(
                {
                    "clean_header": [""],
                }
            ),
            False,
        ),
    ],
)
def test_transform(df_emails, expected_result):
    """Unit test of the transform() method."""

    # Copy for later load/save test
    df_copy = df_emails.copy()

    # Instantiate manually a detector
    detector = ReplyDetector(
        name="reply",
        header_column="clean_header",
    )

    # Get result column name
    res_col = detector.result_column

    # Apply the detector on data
    df_emails = detector.transform(df_emails)

    # Verify result
    result = df_emails[res_col][0]
    assert result == expected_result


@pytest.mark.parametrize(
    "df_emails, expected_result, expected_debug_info",
    [
        (
            DataFrame(
                {
                    "clean_header": ["Re: Suivi de dossier"],
                }
            ),
            True,
            {
                "ReplyRegex": {
                    "match_result": True,
                    "negative_match_data": {},
                    "neutral_match_data": {},
                    "positive_match_data": {"DEFAULT": [{"match_text": "re:", "start": 0, "stop": 3}]},
                },
                "reply_text": "re: suivi de dossier",
            },
        ),
    ],
)
def test_transform_debug_mode(df_emails, expected_result, expected_debug_info):
    """Unit test of the debug mode."""

    # Copy for later load/save test
    df_copy = df_emails.copy()

    # Instanciate manually a detector
    detector = ReplyDetector(
        name="reply",
        header_column="clean_header",
    )

    # Get column names
    res_col = detector.result_column
    debug_dict_col = detector.debug_dict_col

    # Transform data
    df_emails.debug = True
    df_emails = detector.transform(df_emails)

    # Collect results
    result = df_emails[res_col].iloc[0]
    debug_result = df_emails[debug_dict_col].iloc[0]

    # Test result
    assert result == expected_result
    assert debug_result == expected_debug_info


@pytest.mark.parametrize(
    "df, expected_result",
    [
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": ["Re: Suivi de dossier"],
                    "body": ["Bonjour,\nle traitement de ma demande est deplorable.\nje suis tres en colere.\n"],
                }
            ),
            True,
        ),
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": ["re: Envoi d'un document de la Société Imaginaire"],
                    "body": ["Bonjour,\nLe traitement de ma demande est déplorable.\nJe suis très en colère.\n"],
                }
            ),
            True,
        ),
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": ["te: Virement"],
                    "body": [
                        "Bonjour,\nJe vous confirme l'annulation du rdv du 01/01/2022 "
                        + "à 16h.\nBien cordialement,\nJohn Smith."
                    ],
                }
            ),
            False,
        ),
        (
            pd.DataFrame(
                {
                    "from": ["test@gmail.com"],
                    "header": [""],
                    "body": [
                        "Bonjour,\nJe vous confirme l'annulation du rdv du 01/01/2022 "
                        + "à 16h.\nBien cordialement,\nJohn Smith."
                    ],
                }
            ),
            False,
        ),
    ],
)
def test_pipeline_from_config(df, expected_result):
    """
    Instanciate from a config and test the pipeline.
    """
    # Pipeline config key
    pipeline_key = "reply_pipeline"

    # Create pipeline from config
    pipeline = MelusinePipeline.from_config(config_key=pipeline_key)

    # Apply pipeline on data
    df_transformed = pipeline.transform(df)
    result = df_transformed["reply_result"][0]

    # Check
    assert result == expected_result



import pytest


@pytest.fixture
def conf_normalizer():
    return {
        "form": "NFKD",
        "input_columns": ["text"],
        "lowercase": True,
        "output_columns": ["text"],
    }


@pytest.fixture
def conf_tokenizer():
    return {
        "stopwords": ["le", "les"],
        "tokenizer_regex": '\\w+(?:[\\?\\-\\"_]\\w+)*',
    }


@pytest.fixture
def conf_phraser():
    return {
        "input_columns": ["tokens"],
        "output_columns": ["tokens"],
        "threshold": 10,
        "min_count": 10,
    }



from typing import Callable, List

import pandas as pd

from melusine.base import BaseMelusineDetector


# --8<-- [start:detector]
class MyCustomDetector(BaseMelusineDetector):
    @property
    def transform_methods(self) -> List[Callable]:
        return [self.prepare, self.run]

    def prepare(self, row, debug_mode=False):
        return row

    def run(self, row, debug_mode=False):
        row[self.output_columns[0]] = "12345"
        return row


# --8<-- [end:detector]


def run():
    # --8<-- [start:run]
    df = pd.DataFrame(
        [
            {"input_col": "test1"},
            {"input_col": "test2"},
        ]
    )

    detector = MyCustomDetector(input_columns=["input_col"], output_columns=["output_col"], name="custom")

    df = detector.transform(df)
    # --8<-- [end:run]

    return df



import pandas as pd
import pytest

from melusine.base import MelusineTransformer, TransformError


class FakeClass(MelusineTransformer):
    def __init__(self):
        super().__init__(input_columns="input_col", output_columns="output_col", func=self.my_method)

    def my_method(self, df, debug_mode=False):
        raise ValueError


def test_transform_error():
    df = pd.DataFrame([{"input_col": "test"}])
    instance = FakeClass()

    with pytest.raises(TransformError, match="FakeClass.*my_method.*input_col"):
        _ = instance.transform(df)


def test_from_config_from_key_error():
    """Unit test"""

    with pytest.raises(ValueError):
        _ = MelusineTransformer.from_config(config_key="key", config_dict={"foo": "bar"})


def test_missing_func():
    """Unit test"""
    df = pd.DataFrame([{"a": [1, 2, 3]}])
    transformer = MelusineTransformer(input_columns=["a"], output_columns=["b"])

    with pytest.raises(AttributeError):
        transformer.transform(df)



"""
Melusine transformation can operate on different data structures such as dict or pandas.DataFrame.
Different transformation backends are used to process different data structures.
The ActiveBackend class stores an instance of the activated backend.

Implemented classes: [
    ActiveBackend,
]
"""

import logging
from typing import Callable, List, Optional, Union

from melusine.backend.base_backend import Any, BaseTransformerBackend
from melusine.backend.dict_backend import DictBackend

logger = logging.getLogger(__name__)


class ActiveBackend(BaseTransformerBackend):
    """
    Class storing the active backend used by Melusine.
    """

    PANDAS_BACKEND: str = "pandas"
    DICT_BACKEND: str = "dict"

    def __init__(self) -> None:
        """Init"""
        super().__init__()
        self._backend: Optional[BaseTransformerBackend] = None

    @property
    def backend(self) -> BaseTransformerBackend:
        """Backend attribute getter"""
        if self._backend is None:
            raise AttributeError("'_backend' attribute is None")

        else:
            return self._backend

    def reset(self, new_backend: Union[BaseTransformerBackend, str] = PANDAS_BACKEND) -> None:
        """
        Method to switch from current backend to specified backend.

        Parameters
        ----------
        new_backend: Union[BaseTransformerBackend, str]
            New backend to be used
        """

        if isinstance(new_backend, BaseTransformerBackend):
            self._backend = new_backend

        elif new_backend == self.PANDAS_BACKEND:
            # Importing in local scope to prevent hard dependencies
            from melusine.backend.pandas_backend import PandasBackend

            self._backend = PandasBackend()

        elif new_backend == self.DICT_BACKEND:
            self._backend = DictBackend()

        else:
            raise ValueError(f"Backend {new_backend} is not supported")

        logger.info(f"Using backend '{new_backend}' for Data transformations")

    def apply_transform(
        self,
        data: Any,
        func: Callable,
        output_columns: Optional[List[str]] = None,
        input_columns: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Method to apply a transform on a Dataset using current backend.

        Parameters
        ----------
        data: Dataset
            Data to be transformed
        func: Callable
            Transform function to apply to the input data
        output_columns: Sequence[str]
            List of output columns
        input_columns: Sequence[str]
            List of input columns
        kwargs

        Returns
        -------
        _: Dataset
            Transformed data
        """
        return self.backend.apply_transform(
            data=data,
            func=func,
            output_columns=output_columns,
            input_columns=input_columns,
            **kwargs,
        )

    def copy(self, data: Any, fields: Optional[List[str]] = None) -> Any:
        """
        Method to make a copy of the input dataset.

        Parameters
        ----------
        data: Dataset
            MelusineDataset object
        fields: List[str]
            List of fields to include in the copy (by default copy all fields)

        Returns
        -------
        _: Dataset
            Copy of original object
        """
        return self.backend.copy(data, fields=fields)

    def get_fields(self, data: Any) -> List[str]:
        """
        Method to get the list of fields available in the input dataset.

        Parameters
        ----------
        data: Dataset
            MelusineDataset object

        Returns
        -------
        _: List[str]
            List of dataset fields
        """
        return self.backend.get_fields(data=data)

    def add_fields(self, left: Any, right: Any, fields: Optional[List[str]] = None) -> Any:
        """
        Method to add fields from the right object to the left object

        Parameters
        ----------
        left: Dataset
            MelusineDataset object
        right: Dataset
            Melusine Dataset object
        fields: List[str]
            List of fields to be added

        Returns
        -------
        _: Dataset
            Left object with added fields
        """
        return self.backend.add_fields(left=left, right=right, fields=fields)

    def check_debug_flag(self, data: Any) -> bool:
        """
        Method to check if debug_mode is activated.

        Parameters
        ----------
        data: Dataset
            MelusineDataset object

        Returns
        -------
        _: bool
            True if debug mode is activated
        """
        return self.backend.check_debug_flag(data=data)

    def setup_debug_dict(self, data: Any, dict_name: str) -> Any:
        """
        Method to check if debug_mode is activated

        Parameters
        ----------
        data: Dataset
            MelusineDataset object
        dict_name: str
            Name of the debug dict field to be added

        Returns
        -------
        _: Dataset
            MelusineDataset object
        """
        return self.backend.setup_debug_dict(data=data, dict_name=dict_name)


# Instantiate the default backend
backend = ActiveBackend()
backend.reset()



import re

import pytest

from melusine.message import Message
from melusine.processors import BaseContentTagger, ContentTagger, Tag


def test_content_tagger():
    # Text segments (= individual messages in an email conversation)
    text_segments = [
        "Envoye de mon iphone",
        ("Bonjour Mme X,\nSuite a blh blah blah\n" "Bien cordialement\nJane Dupond\n" "(See attached file: flex.jpg)"),
        (
            "Bonjour,\nVeuillez trouver ci-joint blah\n"
            "Merci d'avance,\nCordialement,\n"
            "Toute modification, edition, utilisation ou diffusion non autorisee est interdite"
        ),
    ]

    # Expected tags
    expected_tags = [
        [
            ("FOOTER", "Envoye de mon iphone"),
        ],
        [
            ("HELLO", "Bonjour Mme X,"),
            ("BODY", "Suite a blh blah blah"),
            ("GREETINGS", "Bien cordialement"),
            ("SIGNATURE_NAME", "Jane Dupond"),
            ("PJ", "(See attached file: flex.jpg)"),
        ],
        [
            ("HELLO", "Bonjour,"),
            ("BODY", "Veuillez trouver ci-joint blah"),
            ("THANKS", "Merci d'avance,"),
            ("GREETINGS", "Cordialement,"),
            (
                "FOOTER",
                "Toute modification, edition, utilisation ou diffusion non autorisee est interdite",
            ),
        ],
    ]

    # Mock the output of a Segmenter (List of Message object)
    messages = [Message(text=segment) for segment in text_segments]

    # Instantiate and apply the Tagger
    tagger = ContentTagger()
    output_messages = tagger.tag_email(messages)

    # Test output tags
    output_tags = [x.tags for x in output_messages]
    assert output_tags == expected_tags


def test_tag_null_message():
    messages = None

    # Instantiate and apply the Tagger
    tagger = ContentTagger()
    output_messages = tagger.tag_email(messages)

    assert output_messages is None


@pytest.mark.parametrize(
    "text, expected_parts",
    [
        (
            "Bonjour, merci pour votre message!\nComment-allez vous?! Je suis satisfait!!!\n"
            "Bien cordialement\n\n\n\nJane Dupond\n",
            [
                "Bonjour,",
                "merci pour votre message!",
                "Comment-allez vous?!",
                "Je suis satisfait!!!",
                "Bien cordialement",
                "Jane Dupond",
            ],
        ),
    ],
)
def test_content_tagger_split_text(text, expected_parts):
    # Instantiate and apply the Tagger
    tagger = ContentTagger()
    output_parts = tagger.split_text(text)

    assert output_parts == expected_parts


@pytest.mark.parametrize(
    "text, expected_tags",
    [
        (
            "Bonjour Mme X,\nSuite a blh blah blah.\n"
            "Bien cordialement\nJane Dupond\n"
            "(See attached file: flex.jpg)",
            [
                ("HELLO", "Bonjour Mme X,"),
                ("BODY", "Suite a blh blah blah."),
                ("GREETINGS", "Bien cordialement"),
                ("SIGNATURE_NAME", "Jane Dupond"),
                ("PJ", "(See attached file: flex.jpg)"),
            ],
        ),
        (
            "Bonjour, je confirme le rdv. Cordialement, John Smith",
            [
                ("HELLO", "Bonjour,"),
                ("BODY", "je confirme le rdv."),
                ("GREETINGS", "Cordialement, John Smith"),
            ],
        ),
        (
            (
                "Bonjour,\nSuite a notre intervention du 16.02.22 , un taux d'humidité de 50% a été relevé.\n"
                "Cordialement.\n177, rue de la fée - 75000 Paris.\n"
                "Horaires : du lundi au jeudi de 08h00 à 16h30 et le vendredi de 08h00 à 16h00.\n"
                "Tel : 01.45.53.11.33"
            ),
            [
                ("HELLO", "Bonjour,"),
                ("BODY", "Suite a notre intervention du 16.02.22 , un taux d'humidité de 50% a été relevé."),
                ("GREETINGS", "Cordialement."),
                ("SIGNATURE", "177, rue de la fée - 75000 Paris."),
                ("BODY", "Horaires : du lundi au jeudi de 08h00 à 16h30 et le vendredi de 08h00 à 16h00."),
                ("SIGNATURE", "Tel : 01.45.53.11.33"),
            ],
        ),
        (
            (
                "bonjour\n"
                "15 jours après les premières réparations, un défaut a été détecté. "
                "Bien à vous\n"
                "Britney Spears"
            ),
            [
                ("HELLO", "bonjour"),
                ("BODY", "15 jours après les premières réparations, un défaut a été détecté."),
                ("GREETINGS", "Bien à vous"),
                ("SIGNATURE_NAME", "Britney Spears"),
            ],
        ),
        (
            (
                "Bonjour monsieur Smith\n"
                "merci. Bien à vous\n"
                "Britney Spears\n"
                "22 hollywood boulevard\n"
                "79000 Niort\n"
            ),
            [
                ("HELLO", "Bonjour monsieur Smith"),
                ("THANKS", "merci."),
                ("GREETINGS", "Bien à vous"),
                ("SIGNATURE_NAME", "Britney Spears"),
                ("SIGNATURE", "22 hollywood boulevard"),
                ("SIGNATURE", "79000 Niort"),
            ],
        ),
        (
            (
                "Merci de me faire suivre les docs à ma nouvelle adresse qui est 0 rue du parc, 75000 Paris. "
                "Merci d'avance. \nAcceptez notre salutation,"
            ),
            [
                ("BODY", "Merci de me faire suivre les docs à ma nouvelle adresse qui est 0 rue du parc, 75000 Paris."),
                ("THANKS", "Merci d'avance."),
                ("GREETINGS", "Acceptez notre salutation,"),
            ],
        ),
        (
            (
                "Bonjour\n"
                "Je vous relance concernant ma télévision avec le devis en PJ.\n"
                "Désolé pour la qualité.\n"
                "Je l'ai envoyé à partir de mon ordi.\n"
                "Excellente journée à vous,\n"
                "Bon we\n"
                "Votre bien dévoué\n"
                "amicalement votre\n"
                "Cordiales salutations.\n"
                "Françoise-Bénédicte Dupond\n"
                "Envoyé à partir de \nCourrier \npour Windows"
            ),
            [
                ("HELLO", "Bonjour"),
                ("BODY", "Je vous relance concernant ma télévision avec le devis en PJ."),
                ("BODY", "Désolé pour la qualité."),
                ("BODY", "Je l'ai envoyé à partir de mon ordi."),
                ("HELLO", "Excellente journée à vous,"),
                ("HELLO", "Bon we"),
                ("GREETINGS", "Votre bien dévoué"),
                ("GREETINGS", "amicalement votre"),
                ("GREETINGS", "Cordiales salutations."),
                ("SIGNATURE_NAME", "Françoise-Bénédicte Dupond"),
                ("FOOTER", "Envoyé à partir de"),
                ("FOOTER", "Courrier"),
                ("FOOTER", "pour Windows"),
            ],
        ),
        (
            "C'est bien note, merci beaucoup.\nSentiments dévoués.\nTélécharger \nOutlook pour Android",
            [
                ("THANKS", "C'est bien note, merci beaucoup."),
                ("GREETINGS", "Sentiments dévoués."),
                ("FOOTER", "Télécharger"),
                ("FOOTER", "Outlook pour Android"),
            ],
        ),
        (
            "Impeccable, je vous remercie beaucoup pour votre rapidité.\nObtenir\nOutlook pour Android",
            [
                ("THANKS", "Impeccable, je vous remercie beaucoup pour votre rapidité."),
                ("FOOTER", "Obtenir"),
                ("FOOTER", "Outlook pour Android"),
            ],
        ),
        (
            (
                "Cher Monsieur,\nJe vous confirme la bonne réception de votre précédent email.\n"
                "Je vous en remercie.\nBien cordialement,\nJohn Smith"
            ),
            [
                ("HELLO", "Cher Monsieur,"),
                ("BODY", "Je vous confirme la bonne réception de votre précédent email."),
                ("THANKS", "Je vous en remercie."),
                ("GREETINGS", "Bien cordialement,"),
                ("SIGNATURE_NAME", "John Smith"),
            ],
        ),
        (
            (
                "chère madame,\n"
                "URGENT URGENT\n"
                "Merci de me faire suivre les docs à ma nouvelle adresse qui est 0 rue du parc, 75000 Paris. "
                "Merci d'avance. \nRecevez nos salutations,\nVous en souhaitant bonne réception"
            ),
            [
                ("HELLO", "chère madame,"),
                ("BODY", "URGENT URGENT"),
                ("BODY", "Merci de me faire suivre les docs à ma nouvelle adresse qui est 0 rue du parc, 75000 Paris."),
                ("THANKS", "Merci d'avance."),
                ("GREETINGS", "Recevez nos salutations,"),
                ("GREETINGS", "Vous en souhaitant bonne réception"),
            ],
        ),
        pytest.param(
            "Un témoignage sous X\nEnvoyé depuis mon téléphone Orange",
            [
                ("BODY", "Un témoignage sous X"),
                ("FOOTER", "Envoyé depuis mon téléphone Orange"),
            ],
            id="Edge case where a line ends with an isolated character",
        ),
        pytest.param(
            "     ??\n  !??!",
            [
                ("BODY", "??!??!"),
            ],
            id="Edge case where the two first lines are missing word characters",
        ),
        (
            "Bonjour Mme X,\nSuite a blh blah blah.\n"
            "Bien cordialement\nJane Dupond\n"
            "(See attached file: flex.jpg)",
            [
                ("HELLO", "Bonjour Mme X,"),
                ("BODY", "Suite a blh blah blah."),
                ("GREETINGS", "Bien cordialement"),
                ("SIGNATURE_NAME", "Jane Dupond"),
                ("PJ", "(See attached file: flex.jpg)"),
            ],
        ),
        (
            "\nChère Madame\n\nC'est bien noté, merci\nBien reçu\nJ.Smith\n\n",
            [
                ("HELLO", "Chère Madame"),
                ("THANKS", "C'est bien noté, merci"),
                ("BODY", "Bien reçu"),
                ("SIGNATURE_NAME", "J.Smith"),
            ],
        ),
        (
            "\nBonjour Monsieur, ceci n'est pas un hello\nBonne fin de journee\nsalutations",
            [
                ("BODY", "Bonjour Monsieur, ceci n'est pas un hello"),
                ("HELLO", "Bonne fin de journee"),
                ("GREETINGS", "salutations"),
            ],
        ),
        (
            "\nBonjour Monsieur Stanislas von den hoeggenboord\n\nbien à toi\nJ.  Smith\nChargé de clientèle",
            [
                ("HELLO", "Bonjour Monsieur Stanislas von den hoeggenboord"),
                ("GREETINGS", "bien à toi"),
                ("SIGNATURE_NAME", "J. Smith"),
                ("SIGNATURE", "Chargé de clientèle"),
            ],
        ),
        (
            (
                "\n1 rdv à 18h\n\n2 ème message laissé à la locataire\n3je m'en vais au bois\n"
                "4 allée des iris\n 5bis rue Patrick Sebastien\n6-8 cours mirabeau\n 7 ter place du dahu\n"
                "8 de la rue très longue qui ne doit pas être taggée signature"
            ),
            [
                ("BODY", "1 rdv à 18h"),
                ("BODY", "2 ème message laissé à la locataire"),
                ("BODY", "3je m'en vais au bois"),
                ("SIGNATURE", "4 allée des iris"),
                ("SIGNATURE", "5bis rue Patrick Sebastien"),
                ("SIGNATURE", "6-8 cours mirabeau"),
                ("SIGNATURE", "7 ter place du dahu"),
                ("BODY", "8 de la rue très longue qui ne doit pas être taggée signature"),
            ],
        ),
        (
            (
                "à L'attention de M Bob,\n"
                "Bonjour,\n"
                "Je vous informe que je vais accepter la proposition de L , à savoir le paiement d'une indemnité forfaitaire de résiliation  du CCMI de  4000 € TTC pour clore cette affaire.\n"
                "Cordialement.\n"
                "Bob Smith"
            ),
            [
                ("FOOTER", "à L'attention de M Bob,"),
                ("HELLO", "Bonjour,"),
                (
                    "BODY",
                    "Je vous informe que je vais accepter la proposition de L , à savoir le paiement d'une indemnité forfaitaire de résiliation du CCMI de 4000 € TTC pour clore cette affaire.",
                ),
                ("GREETINGS", "Cordialement."),
                ("SIGNATURE_NAME", "Bob Smith"),
            ],
        ),
        (
            (
                "Monsieur Bob Smith\n"
                "Adresse mail : BobSmith90@gmail.com\n"
                "Lucy Ange\n\n"
                "Bonjour Monsieur,\n"
                "Suite à notre entretien téléphonique de ce matin, et au message que vous m'avez envoyé sur ma messagerie, je voudrais effectuer la réparation du véhicule Renault Twingo dans un garage partenaire de la Maif situé, si c'est possible.\n"
                "Dans l'attente de votre réponse et en vous remerciant par avance,\n\n\n"
                "Monsieur Bob Smith\n\n\n"
                "Envoyé à partir de\n"
                "Courrier\npour Windows\n\n\n\n"
                "Sans virus.\nwww.avast.com"
            ),
            [
                ("HELLO", "Monsieur Bob Smith"),
                ("SIGNATURE", "Adresse mail : BobSmith90@gmail.com"),
                ("SIGNATURE_NAME", "Lucy Ange"),
                ("HELLO", "Bonjour Monsieur,"),
                (
                    "BODY",
                    "Suite à notre entretien téléphonique de ce matin, et au message que vous m'avez envoyé sur ma messagerie, je voudrais effectuer la réparation du véhicule Renault Twingo dans un garage partenaire de la Maif situé, si c'est possible.",
                ),
                ("BODY", "Dans l'attente de votre réponse et en vous remerciant par avance,"),
                ("HELLO", "Monsieur Bob Smith"),
                ("FOOTER", "Envoyé à partir de"),
                ("FOOTER", "Courrier"),
                ("FOOTER", "pour Windows"),
                ("FOOTER", "Sans virus."),
                ("FOOTER", "www.avast.com"),
            ],
        ),
        (
            (
                "Bob Smith\n\n\n"
                "A l’attention de Madame Lucy Ange,\n\n\n\n\n\n"
                "Bonjour Madame Ange,\n\n\n\n\n\n\n\n\n"
                "J’espère que vous allez bien.\n\n\n\n\n\n"
                "Pour faire suite à mon mail du 21 février 2023, je me permets de revenir vers vous pour avoir votre avis sur le devis que j’ai demandé auprès d’un enquêteur.\n\n\n\n"
                "Voici son retour :\n\n\n\n\n\n"
                "Qu’en pensez-vous svp ?\n\n\n\n\n\n"
                "Je reste à votre disposition pour tout complément d’information et vous remercie de l’intérêt que vous porterez à ma demande,\n\n\n\n\n\n"
                "Bien Cordialement,\n\n\n\n\n\n"
                "Bob Smith\n\n\n"
                "Tél. 06.83.22.95.94"
            ),
            [
                ("SIGNATURE_NAME", "Bob Smith"),
                ("FOOTER", "A l’attention de Madame Lucy Ange,"),
                ("HELLO", "Bonjour Madame Ange,"),
                ("BODY", "J’espère que vous allez bien."),
                (
                    "BODY",
                    "Pour faire suite à mon mail du 21 février 2023, je me permets de revenir vers vous pour avoir votre avis sur le devis que j’ai demandé auprès d’un enquêteur.",
                ),
                ("BODY", "Voici son retour :"),
                ("BODY", "Qu’en pensez-vous svp ?"),
                (
                    "BODY",
                    "Je reste à votre disposition pour tout complément d’information et vous remercie de l’intérêt que vous porterez à ma demande,",
                ),
                ("GREETINGS", "Bien Cordialement,"),
                ("SIGNATURE_NAME", "Bob Smith"),
                ("SIGNATURE", "Tél."),
                ("SIGNATURE", "06.83.22.95.94"),
            ],
        ),
        pytest.param(
            (
                "cordialement\nContact e-mail\n\n\nContact téléphone\n\n01 23 45 67 89 / abcabc@hotmail.fr\n"
                "Torroella de Montgri, le 5 avril 2023\nLes formats de fichiers acceptés sont : PDF, DOC, DOCX, JPEG, "
                "JPG, TIFF, TXT, ODT, XLS, XLSX\nTout autre format de fichiers ne sera pas transmis au dossier"
            ),
            [
                ("GREETINGS", "cordialement"),
                ("SIGNATURE", "Contact e-mail"),
                ("SIGNATURE", "Contact téléphone"),
                ("SIGNATURE", "01 23 45 67 89 / abcabc@hotmail.fr"),
                ("SIGNATURE", "Torroella de Montgri, le 5 avril 2023"),
                (
                    "FOOTER",
                    "Les formats de fichiers acceptés sont : PDF, DOC, DOCX, JPEG, JPG, TIFF, TXT, ODT, XLS, XLSX",
                ),
                ("FOOTER", "Tout autre format de fichiers ne sera pas transmis au dossier"),
            ],
            id="diverse_signature_patterns",
        ),
        pytest.param(
            (
                "bonjour\nmon body\nJ. Smith\n\n01 23 45 67 89\nSecrétaire en charge des avions\n"
                "Business Analyst – Tribu Sinistres – Squad Flux Entrants\n"
                "Société nationale des chemins de fer\nConseiller MAIF\nGestionnaire sinistre - C99G\n"
                "Service des lettres anonymes\nTechnicienne de gestion - EQUIPE ABC\n"
            ),
            [
                ("HELLO", "bonjour"),
                ("BODY", "mon body"),
                ("SIGNATURE_NAME", "J. Smith"),
                ("SIGNATURE", "01 23 45 67 89"),
                ("SIGNATURE", "Secrétaire en charge des avions"),
                ("SIGNATURE", "Business Analyst – Tribu Sinistres – Squad Flux Entrants"),
                ("SIGNATURE", "Société nationale des chemins de fer"),
                ("SIGNATURE", "Conseiller MAIF"),
                ("SIGNATURE", "Gestionnaire sinistre - C99G"),
                ("SIGNATURE", "Service des lettres anonymes"),
                ("SIGNATURE", "Technicienne de gestion - EQUIPE ABC"),
            ],
            id="signature_jobs",
        ),
        pytest.param(
            (
                "bonjour\nmon body\nCordialement\n\n"
                "analyste -------------------------------------- test test test test test test test\n"
            ),
            [
                ("HELLO", "bonjour"),
                ("BODY", "mon body"),
                ("GREETINGS", "Cordialement"),
                ("BODY", "analyste -------------------------------------- test test test test test test test"),
            ],
            id="check_catastrophic_backtracking",
        ),
    ],
)
def test_tag_text_generic(text, expected_tags):
    # Instantiate and apply the Tagger
    tagger = ContentTagger()
    output_tags = tagger.tag_text(text)
    # Test output tags
    assert output_tags == expected_tags


@pytest.mark.parametrize(
    "text, expected_tags",
    [
        pytest.param(
            (
                "Merci\n"
                "Je vous remercie\n"
                "Merci d'avance\n"
                "Je vous remercie par avance\n"
                "Vous en remerciant par avance.\n"
            ),
            [
                ("THANKS", "Merci"),
                ("THANKS", "Je vous remercie"),
                ("THANKS", "Merci d'avance"),
                ("THANKS", "Je vous remercie par avance"),
                ("THANKS", "Vous en remerciant par avance."),
            ],
            id="french thanks patterns",
        ),
    ],
)
def test_tag_text_french(text, expected_tags):
    # Instantiate and apply the Tagger
    tagger = ContentTagger()
    output_tags = tagger.tag_text(text)
    # Test output tags
    assert output_tags == expected_tags


@pytest.mark.parametrize(
    "text, expected_tags",
    [
        pytest.param(
            (
                "Thank you so much\n"
                "thanks\n"
                "thx Joanna\n"
                "thanks but you forgot bla\n"
                "Thx however I still need the document\n"
            ),
            [
                ("THANKS", "Thank you so much"),
                ("THANKS", "thanks"),
                ("THANKS", "thx Joanna"),
                ("BODY", "thanks but you forgot bla"),
                ("BODY", "Thx however I still need the document"),
            ],
            id="english thanks patterns",
        ),
        pytest.param(
            (
                "Best\n"
                "warm Wishes\n"
                "regards\n"
                "best regards\n"
                "cheers\n"
                "yours\n"
                "yours truly\n"
                "Sincerely\n"
                "see you soon\n"
                "Speak to you soon\n"
                "talk soon\n"
                "Take care\n"
                "Catch you later\n"
                "Have a fantastic day\n"
                "Looking forward to your reply\n"
                "I am looking forward to hearing from you\n"
                "Hoping to hear from you\n"
            ),
            [
                ("GREETINGS", "Best"),
                ("GREETINGS", "warm Wishes"),
                ("GREETINGS", "regards"),
                ("GREETINGS", "best regards"),
                ("GREETINGS", "cheers"),
                ("GREETINGS", "yours"),
                ("GREETINGS", "yours truly"),
                ("GREETINGS", "Sincerely"),
                ("GREETINGS", "see you soon"),
                ("GREETINGS", "Speak to you soon"),
                ("GREETINGS", "talk soon"),
                ("GREETINGS", "Take care"),
                ("GREETINGS", "Catch you later"),
                ("GREETINGS", "Have a fantastic day"),
                ("GREETINGS", "Looking forward to your reply"),
                ("GREETINGS", "I am looking forward to hearing from you"),
                ("GREETINGS", "Hoping to hear from you"),
            ],
            id="english greetings",
        ),
        pytest.param(
            (
                "Hello John\n"
                "hi\n"
                "Hi there\n"
                "good to hear from you\n"
                "it is good to hear from you\n"
                "I hope you are having a great week\n"
                "how are you doing\n"
                "how are you positioned about the matter\n"
                "i hope you are doing well\n"
                "Good Morning Joanna\n"
                "good Afternoon\n"
                "Dear Jacky\n"
                "Sir\n"
                "Dear Madam\n"
                "Dear Mr\n"
                "Dear Ms.\n"
                "Dear miss\n"
                "Dear mrs.\n"
                "Dear sir or madam\n"
                "To whom it may concern\n"
            ),
            [
                ("HELLO", "Hello John"),
                ("HELLO", "hi"),
                ("HELLO", "Hi there"),
                ("HELLO", "good to hear from you"),
                ("HELLO", "it is good to hear from you"),
                ("HELLO", "I hope you are having a great week"),
                ("HELLO", "how are you doing"),
                ("BODY", "how are you positioned about the matter"),
                ("HELLO", "i hope you are doing well"),
                ("HELLO", "Good Morning Joanna"),
                ("HELLO", "good Afternoon"),
                ("HELLO", "Dear Jacky"),
                ("HELLO", "Sir"),
                ("HELLO", "Dear Madam"),
                ("HELLO", "Dear Mr"),
                ("HELLO", "Dear Ms."),
                ("HELLO", "Dear miss"),
                ("HELLO", "Dear mrs."),
                ("HELLO", "Dear sir or madam"),
                ("HELLO", "To whom it may concern"),
            ],
            id="english hello",
        ),
        pytest.param(
            (
                "VP of Data Science\n"
                "Chief of staff\n"
                "CTO at TestMelusine\n"
                "CEOABC test\n"
                "Lead business developer\n"
            ),
            [
                ("SIGNATURE", "VP of Data Science"),
                ("SIGNATURE", "Chief of staff"),
                ("SIGNATURE", "CTO at TestMelusine"),
                ("BODY", "CEOABC test"),
                ("SIGNATURE", "Lead business developer"),
            ],
            id="english job signature patterns",
        ),
        pytest.param(
            (
                "9 downing street\n"
                "4-6 Beverly Hill\n"
                "4 Abbey road W24RA\n"
                "3 Ocean Rd.\n"
                "5th avenue\n"
                "221b Baker St.\n"
                "6bis River ln.\n"
                "7 Winter lane\n"
            ),
            [
                ("SIGNATURE", "9 downing street"),
                ("SIGNATURE", "4-6 Beverly Hill"),
                ("SIGNATURE", "4 Abbey road W24RA"),
                ("SIGNATURE", "3 Ocean Rd."),
                ("SIGNATURE", "5th avenue"),
                ("SIGNATURE", "221b Baker St."),
                ("SIGNATURE", "6bis River ln."),
                ("SIGNATURE", "7 Winter lane"),
            ],
            id="english adsress signature patterns",
        ),
    ],
)
def test_tag_text_english(text, expected_tags):
    # Instantiate and apply the Tagger
    tagger = ContentTagger()
    output_tags = tagger.tag_text(text)
    # Test output tags
    assert output_tags == expected_tags


def test_tag_list():
    # Limit tags to "HELLO" and the default tag ("BODY")
    tag_list = ["HELLO"]

    # Text segment (= individual message in an email conversation)
    text = "bonjour\nblah blah blah\nmerci\ncordialement"

    # Expected tags
    expected_tags = [
        ("HELLO", "bonjour"),
        ("BODY", "blah blah blah"),
        ("BODY", "merci"),
        ("BODY", "cordialement"),
    ]

    # Instantiate and apply the Tagger
    tagger = ContentTagger(tag_list=tag_list)
    output_tags = tagger.tag_text(text)

    # Test output tags
    assert expected_tags == output_tags


def test_undefined_tag():
    unknown_tag = "UNKNOWN_TAG"

    # Setup an unknown tag
    tag_list = [unknown_tag]

    # Instantiate Tagger
    with pytest.raises(ValueError, match=rf".*{unknown_tag}.*"):
        _ = ContentTagger(tag_list=tag_list)


def test_unsupported_type():
    class MyClass(ContentTagger):
        """Test class"""

        @Tag
        def TEST_TAG(self):
            """Test method"""
            return 3.3

    with pytest.raises(ValueError, match="supported types"):
        _ = MyClass()


def test_compiled_pattern():
    class MyClass(ContentTagger):
        """Test class"""

        @Tag
        def TEST_TAG(self):
            """Test method"""
            return re.compile(r"cool_pattern")

    tagger = MyClass()
    subtext, tag, match = tagger("cool_pattern is what I am looking for")[0]

    # Check tag result
    assert tag == "TEST_TAG"


def test_str_pattern():
    class MyClass(ContentTagger):
        """Test class"""

        @Tag
        def TEST_TAG(self):
            """Test method"""
            return r"cool_pattern"

    tagger = MyClass()
    subtext, tag, match = tagger("cool_pattern is what I am looking for")[0]

    # Check tag result
    assert tag == "TEST_TAG"


def test_malformed_regex():
    from melusine.processors import Tag

    malformed_regex = r"[*."

    # Create a tagger containing an ill defined Tag (malformed regex)
    class CustomTagger(ContentTagger):
        """Test class"""

        @Tag
        def HELLO(self):
            """Test method"""
            return malformed_regex

    # Instantiate Tagger
    with pytest.raises(ValueError, match=rf"Invalid regex"):
        _ = CustomTagger()


def test_direct_tagging():
    tagger = ContentTagger()
    match = tagger["HELLO"].match("Bonjour")

    assert bool(match)


def test_call_method():
    tagger = ContentTagger()

    match_list = tagger("Bonjour a tous")
    subtext, tag, regex = match_list[0]

    assert tag == "HELLO"


@pytest.mark.parametrize(
    "text, n_words, word_character_only, expected_match",
    [
        pytest.param("Hello you", 4, False, True, id="4 words match"),
        pytest.param("Hello how are you today", 4, False, False, id="4 words no match"),
        pytest.param("Hello! you?", 4, False, True, id="4 words match with special characters"),
        pytest.param(
            "Hello! you?", 4, True, False, id="4 words match with special characters (word character only True)"
        ),
    ],
)
def test_word_blocks(text, n_words, word_character_only, expected_match):
    regex = BaseContentTagger.word_block(n_words, word_character_only=word_character_only)

    search_regex = r"^" + regex + r"$"
    match = bool(re.search(search_regex, text))
    assert match == expected_match



from typing import Dict, List, Optional, Union

from melusine.base import MelusineRegex


class TransferRegex(MelusineRegex):
    """
    Detect transfer patterns in headers such as "tr:".
    """

    @property
    def positive(self) -> Union[str, Dict[str, str]]:
        """
        Define regex patterns required to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return r"^(tr:|fwd :|tr :|fwd:)"

    @property
    def neutral(self) -> Optional[Union[str, Dict[str, str]]]:
        """
        Define regex patterns to be ignored when running detection.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return None

    @property
    def negative(self) -> Optional[Union[str, Dict[str, str]]]:
        """
        Define regex patterns prohibited to activate the MelusineRegex.

        Returns:
            _: Regex pattern or dict of regex patterns.
        """
        return None

    @property
    def match_list(self) -> List[str]:
        """
        List of texts that should activate the MelusineRegex.

        Returns:
            _: List of texts.
        """
        return [
            "tr: notre discussion",
            "tr : bonjour",
            "Tr : compte rendu",
            "fwd: rdv du 01/01/2001",
            "Fwd: Votre déclaration",
        ]

    @property
    def no_match_list(self) -> List[str]:
        """
        List of texts that should NOT activate the MelusineRegex.

        Returns:
            _: List of texts.
        """
        return ["transfert d'argent", "re: tr: message du jour" "re:Fwd important notice"]



import pytest

from melusine.testing import assert_pipeline_results

from .test_emails_fixtures import testcase


# The test_message fixture sequentially takes the value of the
# test_cases defined in melusine_code/tests/fixtures/test_emails_fixtures.py
@pytest.mark.usefixtures("use_dict_backend")
def test_pipeline_steps(testcase):
    # Run pipeline tests
    pipeline_name = testcase.pop("pipeline")
    assert_pipeline_results(testcase, pipeline_name)
